import pytest

from NekUpload.NekData.composite import CompositeDefinition,CompositeValidationException
from NekUpload.NekData.data_type import Elements

@pytest.mark.unit
@pytest.mark.parametrize("description,composite_definition,expected_shape,expected_count,expected_id_set",[
    ("Test Edge","E[1]",Elements.EDGE,1,{1}),
    ("Test Face","F[2]",Elements.FACE,1,{2}),
    ("Test Face with Leading Edge Whitespace","       F[2]",Elements.FACE,1,{2}),
    ("Test COMPOSITE","C[2]",Elements.COMPOSITE,1,{2}),
    ("Test TRI","T[2-5]",Elements.TRI,4,{2,3,4,5}),
    ("Test QUAD","T[2- 5]",Elements.TRI,4,{2,3,4,5}),
    ("Test TET","A[2-5,8,10]",Elements.TET,6,{2,3,4,5,8,10}),
    ("Test PYR","P[2-5,8,10,99-100]",Elements.PYR,8,{2,3,4,5,8,10,99,100}),
    ("Test PRISM","R[2-5,8,10   ,   99-  100  ]    ",Elements.PRISM,8,{2,3,4,5,8,10,99,100}),
    ("Test HEX","H[2-5,8,10   ,   99-  100  ]    ",Elements.HEX,8,{2,3,4,5,8,10,99,100}),
])
def test_valid_composite_definition(description:str,composite_definition:str,
                                    expected_shape:Elements,expected_count:int,
                                    expected_id_set:set[int]):
    
    composite = CompositeDefinition(composite_definition)
    
    assert composite.count == expected_count, f"{description}: Expected count {expected_count}, but got {composite.count}"
    assert composite.element == expected_shape, f"{description}: Expected shape {expected_shape}, but got {composite.element}"
    assert composite.composite_ids == expected_id_set, f"{description}: Expected ID set {expected_id_set}, but got {composite.composite_ids}"

@pytest.mark.unit
@pytest.mark.parametrize("description,composite_definition",[
    ("Invalid composite id","H[2-5,8,10 , 99- 100A  ]"),
    ("Invalid element identifier","X[2-5,8,10 , 99- 100  ]")
])
def test_undefined_composite_definition(description:str,composite_definition:str):
    with pytest.raises(CompositeValidationException):
        CompositeDefinition(composite_definition)
        pytest.fail(f"Test {description} was not rejected")

@pytest.mark.unit
@pytest.mark.parametrize("composite1,composite2", [
    ("E[1]","E[1]"),
    ("E[1-100]","E[1-100]"),
    ("E[1-3,6]","E[1-3,6]"),
    ("E[1-3,6]","E[6,1-3]"),
])
def test_same(composite1:str,composite2:str):
    item1 = CompositeDefinition(composite1)
    item2 = CompositeDefinition(composite2)

    assert item1 == item2, f"Expected {composite1} to be equal to {composite2}, but they are not."

@pytest.mark.unit
@pytest.mark.parametrize("composite1,composite2", [
    ("E[1]","E[2]"),
    ("E[1-100]","E[5]"),
    ("F[1-3,6]","E[1-3,6]"),
])
def test_diff(composite1:str,composite2:str):
    item1 = CompositeDefinition(composite1)
    item2 = CompositeDefinition(composite2)

    assert item1 != item2, f"Expected {composite1} to not be equal to {composite2}, but they are."