import pytest
from NekUpload.NekData.data_type import Elements,BasisType,IntegrationPoint
import NekUpload.NekData.expansions as nd

###############################################################################
# Expansion Definition functionality
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,expected_coeffs",[
    (Elements.SEG,5,5),
    (Elements.QUAD,5,25),
    (Elements.TRI,5,15),
    (Elements.HEX,5,125),
    (Elements.PRISM,5,75),
    (Elements.PYR,5,55)
])
def test_get_uniorder_coefficients(element: Elements, num_modes: int, expected_coeffs: int):
    # Use modified expansion factory as it provides definitions for all element shapes
    expansion_factory = nd.ModifiedExpansionFactory()
    
    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes)
    assert expansion.get_num_coefficients() == expected_coeffs, f"Expected {expected_coeffs} coefficients for element {element}, but got {expansion.get_num_coefficients()}"

@pytest.mark.unit
@pytest.mark.parametrize("expansion,expected_coeff", [
    (nd.ExpansionDefinition(Elements.QUAD, (BasisType.MODIFIED_A, BasisType.MODIFIED_A), (5, 6),
                            (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), (6, 6)), 30),
    (nd.ExpansionDefinition(Elements.TRI, (BasisType.MODIFIED_A, BasisType.MODIFIED_B), (5, 6),
                            (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), (6, 6)), 20),
    (nd.ExpansionDefinition(Elements.HEX, (BasisType.MODIFIED_A, BasisType.MODIFIED_A,BasisType.MODIFIED_A), (5,6,7),
                            (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
                            (6, 6,6)), 210),
    (nd.ExpansionDefinition(Elements.TET, (BasisType.MODIFIED_A, BasisType.MODIFIED_A,BasisType.MODIFIED_A), (5,6,7),
                            (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
                            (6, 6,6)), 75),
    (nd.ExpansionDefinition(Elements.PYR, (BasisType.MODIFIED_A, BasisType.MODIFIED_A,BasisType.MODIFIED_A), (5,6,7),
                            (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
                            (6, 6,6)), 115),
    (nd.ExpansionDefinition(Elements.PRISM, (BasisType.MODIFIED_A, BasisType.MODIFIED_A,BasisType.MODIFIED_A), (5,6,7),
                            (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
                            (6, 6,6)), 150),
])
def test_get_polyorder_coefficients(expansion: nd.ExpansionDefinition, expected_coeff: int):
    assert expansion.get_num_coefficients() == expected_coeff, f"Expected {expected_coeff} coefficients, but got {expansion.get_num_coefficients()}"

@pytest.mark.unit
@pytest.mark.parametrize("description,element,basis_type,num_modes,integration_point_type,num_points", [
    #Test basic dimensionality issues
    ("1D-wrong-dim-modes",Elements.SEG, (BasisType.MODIFIED_A,), (5,6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,), (6,)),
    ("1D-wrong-dim-points",Elements.SEG, (BasisType.MODIFIED_A,), (5,), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,), (6,6)),
    ("1D-wrong-dim-int-points",Elements.SEG, (BasisType.MODIFIED_A,), (5,), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,), (6,)),
    ("1D-wrong-dim-basis",Elements.SEG, (BasisType.MODIFIED_A,BasisType.MODIFIED_A), (5,), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,), (6,)),

    ("2D-tri-wrong-dim-modes",Elements.TRI, (BasisType.MODIFIED_A, BasisType.MODIFIED_B), (5,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), (6,5)),
    ("2D-tri-wrong-dim-points",Elements.TRI, (BasisType.MODIFIED_A, BasisType.MODIFIED_B), (5,6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), (6,5,5)),
    ("2D-tri-wrong-dim-int-points",Elements.TRI, (BasisType.MODIFIED_A, BasisType.MODIFIED_B), (5,6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,), (6,5)),
    ("2D-tri-wrong-dim-basis",Elements.TRI, (BasisType.MODIFIED_A,), (5,6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), (6,5)),

    ("2D-quad-wrong-dim-modes",Elements.QUAD, (BasisType.MODIFIED_A, BasisType.MODIFIED_A), (5,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), (6,6)),
    ("2D-quad-wrong-dim-points",Elements.QUAD, (BasisType.MODIFIED_A, BasisType.MODIFIED_A), (5,6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), (6,6,6)),
    ("2D-quad-wrong-dim-int-points",Elements.QUAD, (BasisType.MODIFIED_A, BasisType.MODIFIED_A), (5,6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,), (6,6)),
    ("2D-quad-wrong-dim-basis",Elements.QUAD, (BasisType.MODIFIED_A,), (5,6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), (6,6)),

    ("3D-hex-wrong-dim-modes",Elements.HEX, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_A), (5,6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), (6,6,6)),
    ("3D-hex-wrong-dim-points",Elements.HEX, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_A), (5,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), (6,6)),
    ("3D-hex-wrong-dim-int-points",Elements.HEX, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_A), (5,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), (6,6,6)),
    ("3D-hex-wrong-dim-basis",Elements.HEX, (BasisType.MODIFIED_A, BasisType.MODIFIED_A), (5,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), (6,6,6)),

    ("3D-tet-wrong-dim-modes",Elements.TET, (BasisType.MODIFIED_A, BasisType.MODIFIED_B, BasisType.MODIFIED_C), (5,6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), (6,5,5)),
    ("3D-tet-wrong-dim-points",Elements.TET, (BasisType.MODIFIED_A, BasisType.MODIFIED_B, BasisType.MODIFIED_C), (5,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), (6,5)),
    ("3D-tet-wrong-dim-int-points",Elements.TET, (BasisType.MODIFIED_A, BasisType.MODIFIED_B, BasisType.MODIFIED_C), (5,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), (6,5,5)),
    ("3D-tet-wrong-dim-basis",Elements.TET, (BasisType.MODIFIED_A, BasisType.MODIFIED_B), (5,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), (6,5,5)),

    ("3D-pyr-wrong-dim-modes",Elements.PYR, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_PYR_C), (5,6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), (6,6,6)),
    ("3D-pyr-wrong-dim-points",Elements.PYR, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_PYR_C), (5,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), (6,6)),
    ("3D-pyr-wrong-dim-int-points",Elements.PYR, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_PYR_C), (5,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), (6,6,6)),
    ("3D-pyr-wrong-dim-basis",Elements.PYR, (BasisType.MODIFIED_A, BasisType.MODIFIED_A), (5,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), (6,6,6)),

    ("3D-prism-wrong-dim-modes",Elements.PRISM, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_B), (5,6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), (6,6,5)),
    ("3D-prism-wrong-dim-points",Elements.PRISM, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_B), (5,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), (6,6)),
    ("3D-prism-wrong-dim-int-points",Elements.PRISM, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_B), (5,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), (6,6,5)),
    ("3D-prism-wrong-dim-basis",Elements.PRISM, (BasisType.MODIFIED_A, BasisType.MODIFIED_A), (5,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), (6,6,5)),

    #test definition specific issues
    ("2D-tri-first-mode > second-mode",Elements.TRI, (BasisType.MODIFIED_A, BasisType.MODIFIED_B), (6,5), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), (6,5)),
    ("3D-tet-first-mode > third-mode",Elements.TET, (BasisType.MODIFIED_A, BasisType.MODIFIED_B, BasisType.MODIFIED_C), (8,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), (6,5,5)),
    ("3D-tet-second-mode > third-mode",Elements.TET, (BasisType.MODIFIED_A, BasisType.MODIFIED_B, BasisType.MODIFIED_C), (5,8,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), (6,5,5)),
    ("3D-pyr-first-mode > third-mode",Elements.PYR, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_PYR_C), (8,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), (6,6,6)),
    ("3D-pyr-second-mode > third-mode",Elements.PYR, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_PYR_C), (5,8,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), (6,6,6)),
    ("3D-prism-first-mode > third-mode",Elements.PRISM, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_B), (8,6,7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), (6,6,5)),
])
def test_validate_reject_expansion_definition(description: str, element: Elements, 
                                            basis_type: tuple[BasisType, ...], num_modes: tuple[int, ...], 
                                            integration_point_type: tuple[IntegrationPoint, ...], 
                                            num_points: tuple[int, ...]):
    
    # only test rejection as non-rejection implicitly handled already
    with pytest.raises(nd.ExpansionValidationException):
        nd.ExpansionDefinition(element, basis_type, num_modes, integration_point_type, num_points)
        pytest.fail(f"Test {description} failed")

###############################################################################
# Modified Expansion Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.SEG,5,None,(5,),(6,),(IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,),None,(BasisType.MODIFIED_A,)),
    (Elements.TRI,5,None,(5,5),(6,5),(IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0),
        None,(BasisType.MODIFIED_A,BasisType.MODIFIED_B)),
    (Elements.QUAD,5,None,(5,5),(6,6),(IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
        None,(BasisType.MODIFIED_A,BasisType.MODIFIED_A)),
    (Elements.HEX, 5, None, (5, 5, 5), (6, 6, 6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), 
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_A)),
    (Elements.TET, 5, None, (5, 5, 5), (6, 5, 5), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0),
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_B, BasisType.MODIFIED_C)),
    (Elements.PYR, 5, None, (5, 5, 5), (6, 6, 6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0),
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_PYR_C)),
    (Elements.PRISM, 5, None, (5, 5, 5), (6, 6, 5), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), #
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_B))
])
def test_modified_expansion_factory(element:Elements,num_modes:int,fields:tuple[str] | None,
                                    expected_modes:tuple[int],expected_points:tuple[int],
                                    expected_integration_type:tuple[IntegrationPoint],
                                    expected_field:tuple[str] | None,expected_basis:tuple[BasisType]):
    expansion_factory = nd.ModifiedExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element,num_modes,fields) 

    assert_msg_base = f"For ModifiedExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"

###############################################################################
# Modified Quad Plus One Expansion Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.SEG, 5, None, (5,), (7,), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,), None, (BasisType.MODIFIED_A,)),
    (Elements.TRI, 5, None, (5, 5), (7, 6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), 
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_B)),
    (Elements.QUAD, 5, None, (5, 5), (7, 7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), None, 
        (BasisType.MODIFIED_A, BasisType.MODIFIED_A)),
    (Elements.HEX, 5, None, (5, 5, 5), (7, 7, 7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_A)),
    (Elements.TET, 5, None, (5, 5, 5), (7, 6, 6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), 
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_B, BasisType.MODIFIED_C)),
    (Elements.PYR, 5, None, (5, 5, 5), (7, 7, 7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), 
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_PYR_C)),
    (Elements.PRISM, 5, None, (5, 5, 5), (7, 7, 6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), 
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_B))
])
def test_modified_plus_one_expansion_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.ModifiedQuadPlus1ExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For ModifiedQuadPlus1ExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"

###############################################################################
# Modified Quad Plus Two Expansion Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.SEG, 5, None, (5,), (8,), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,), None, (BasisType.MODIFIED_A,)),
    (Elements.TRI, 5, None, (5, 5), (8, 7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), 
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_B)),
    (Elements.QUAD, 5, None, (5, 5), (8, 8), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), None, 
        (BasisType.MODIFIED_A, BasisType.MODIFIED_A)),
    (Elements.HEX, 5, None, (5, 5, 5), (8, 8, 8), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_A)),
    (Elements.TET, 5, None, (5, 5, 5), (8, 7, 7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), 
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_B, BasisType.MODIFIED_C)),
    (Elements.PYR, 5, None, (5, 5, 5), (8, 8, 8), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), 
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_PYR_C)),
    (Elements.PRISM, 5, None, (5, 5, 5), (8, 8, 7), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), 
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_B))
])
def test_modified_plus_two_expansion_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.ModifiedQuadPlus2ExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For ModifiedQuadPlus2ExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"

###############################################################################
# Modified GLL Expansion Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.SEG, 5, None, (5,), (6,), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,), None, (BasisType.MODIFIED_A,)),
    (Elements.TRI, 5, None, (5, 5), (6, 5), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), 
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_B)),
    (Elements.QUAD, 5, None, (5, 5), (6, 6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), None, 
        (BasisType.MODIFIED_A, BasisType.MODIFIED_A)),
    (Elements.HEX, 5, None, (5, 5, 5), (6, 6, 6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_A)),
    (Elements.TET, 5, None, (5, 5, 5), (6, 5, 5), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), 
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_B, BasisType.MODIFIED_C)),
    (Elements.PYR, 5, None, (5, 5, 5), (6, 6, 6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), 
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_PYR_C)),
    (Elements.PRISM, 5, None, (5, 5, 5), (6, 6, 5), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), 
        None, (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_B))
])
def test_modified_GLL_expansion_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.ModifiedGLLRadau10ExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For ModifiedGLLRadau10ExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"

###############################################################################
# GLL Lagrange Expansion Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.SEG, 5, None, (5,), (6,), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,), None, (BasisType.GLL_LAGRANGE,)),
    (Elements.TRI, 5, None, (5, 5), (6, 5), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), 
        None, (BasisType.GLL_LAGRANGE, BasisType.ORTHO_B)),
    (Elements.QUAD, 5, None, (5, 5), (6, 6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), None, 
        (BasisType.GLL_LAGRANGE, BasisType.GLL_LAGRANGE)),
    (Elements.HEX, 5, None, (5, 5, 5), (6, 6, 6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
        None, (BasisType.GLL_LAGRANGE, BasisType.GLL_LAGRANGE, BasisType.GLL_LAGRANGE)),
])
def test_valid_GLL_lagrange_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.GLLLagranageExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For GLLLagranageExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"
    
@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes",[
    (Elements.TET,5),
    (Elements.PYR,5),
    (Elements.PRISM,5),
])
def test_invalid_GLL_lagrange_factory(element:Elements,num_modes:int):
    expansion_factory = nd.GLLLagranageExpansionFactory()
    
    with pytest.raises(nd.ExpansionValidationException):
        expansion_factory.get_expansion(element,num_modes)

###############################################################################
# Gauss Lagrange Expansion Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.SEG, 5, None, (5,), (5,), (IntegrationPoint.GAUSS_GAUSS_LEGENDRE,), None, (BasisType.GAUSS_LAGRANGE,)),
    (Elements.QUAD, 5, None, (5, 5), (5, 5), (IntegrationPoint.GAUSS_GAUSS_LEGENDRE, IntegrationPoint.GAUSS_GAUSS_LEGENDRE), None, 
        (BasisType.GAUSS_LAGRANGE, BasisType.GAUSS_LAGRANGE)),
    (Elements.HEX, 5, None, (5, 5, 5), (5, 5, 5), (IntegrationPoint.GAUSS_GAUSS_LEGENDRE, IntegrationPoint.GAUSS_GAUSS_LEGENDRE, IntegrationPoint.GAUSS_GAUSS_LEGENDRE),
        None, (BasisType.GAUSS_LAGRANGE, BasisType.GAUSS_LAGRANGE, BasisType.GAUSS_LAGRANGE)),
])
def test_valid_gauss_lagrange_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.GaussLagrangeExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For GaussLagrangeExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"
    
@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes",[
    (Elements.TRI,5),
    (Elements.TET,5),
    (Elements.PYR,5),
    (Elements.PRISM,5),
])
def test_invalid_gauss_lagrange_factory(element:Elements,num_modes:int):
    expansion_factory = nd.GaussLagrangeExpansionFactory()
    
    with pytest.raises(nd.ExpansionValidationException):
        expansion_factory.get_expansion(element,num_modes)

###############################################################################
# Orthogonal Expansion Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.SEG, 5, None, (5,), (6,), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,), None, (BasisType.ORTHO_A,)),
    (Elements.TRI, 5, None, (5, 5), (6, 5), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0), 
        None, (BasisType.ORTHO_A, BasisType.ORTHO_B)),
    (Elements.QUAD, 5, None, (5, 5), (6, 6), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), None, 
        (BasisType.ORTHO_A, BasisType.ORTHO_A)),
    (Elements.TET, 5, None, (5, 5, 5), (6, 5, 5), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0), 
        None, (BasisType.ORTHO_A, BasisType.ORTHO_B, BasisType.ORTHO_C)),
])
def test_valid_orthogonal_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.OrthogonalExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For OrthogonalExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"
    
@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes",[
    (Elements.HEX,5),
    (Elements.PYR,5),
    (Elements.PRISM,5),
])
def test_invalid_orthogonal_factory(element:Elements,num_modes:int):
    expansion_factory = nd.OrthogonalExpansionFactory()
    
    with pytest.raises(nd.ExpansionValidationException):
        expansion_factory.get_expansion(element,num_modes)

###############################################################################
# GLL Lagrange SEM Expansion Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.SEG, 5, None, (5,), (5,), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,), None, (BasisType.GLL_LAGRANGE,)),
    (Elements.QUAD, 5, None, (5, 5), (5, 5), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), None, 
        (BasisType.GLL_LAGRANGE, BasisType.GLL_LAGRANGE)),
    (Elements.HEX, 5, None, (5, 5, 5), (5, 5, 5), (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
        None, (BasisType.GLL_LAGRANGE, BasisType.GLL_LAGRANGE, BasisType.GLL_LAGRANGE)),
])
def test_valid_GLL_lagrange_SEM_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.GLLLagrangeSEMExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For GLLLagrangeSEMExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"
    
@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes",[
    (Elements.TRI,5),
    (Elements.TET,5),
    (Elements.PYR,5),
    (Elements.PRISM,5),
])
def test_invalid_GLL_lagrange_SEM_factory(element:Elements,num_modes:int):
    expansion_factory = nd.GLLLagrangeSEMExpansionFactory()
    
    with pytest.raises(nd.ExpansionValidationException):
        expansion_factory.get_expansion(element,num_modes)

###############################################################################
# Fourier Expansion Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.SEG, 5, None, (5,), (5,), (IntegrationPoint.FOURIER_EVENLY_SPACED,), None, (BasisType.FOURIER,)),
    (Elements.QUAD, 5, None, (5, 5), (5, 5), (IntegrationPoint.FOURIER_EVENLY_SPACED, IntegrationPoint.FOURIER_EVENLY_SPACED), None, 
        (BasisType.FOURIER, BasisType.FOURIER)),
    (Elements.HEX, 5, None, (5, 5, 5), (5, 5, 5), (IntegrationPoint.FOURIER_EVENLY_SPACED, IntegrationPoint.FOURIER_EVENLY_SPACED, IntegrationPoint.FOURIER_EVENLY_SPACED),
        None, (BasisType.FOURIER, BasisType.FOURIER, BasisType.FOURIER)),
])
def test_valid_fourier_expansion_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.FourierExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For FourierExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes",[
    (Elements.TRI,5),
    (Elements.TET,5),
    (Elements.PYR,5),
    (Elements.PRISM,5),
])
def test_invalid_fourier_expansion_factory(element:Elements,num_modes:int):
    expansion_factory = nd.FourierExpansionFactory()
    
    with pytest.raises(nd.ExpansionValidationException):
        expansion_factory.get_expansion(element,num_modes)

###############################################################################
# Fourier Single Mode Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.SEG, 5, None, (5,), (5,), (IntegrationPoint.FOURIER_SINGLE_MODE_SPACED,), None, (BasisType.FOURIER_SINGLE_MODE,)),
    (Elements.QUAD, 5, None, (5, 5), (5, 5), (IntegrationPoint.FOURIER_SINGLE_MODE_SPACED, IntegrationPoint.FOURIER_SINGLE_MODE_SPACED), None, 
        (BasisType.FOURIER_SINGLE_MODE, BasisType.FOURIER_SINGLE_MODE)),
    (Elements.HEX, 5, None, (5, 5, 5), (5, 5, 5), (IntegrationPoint.FOURIER_SINGLE_MODE_SPACED, IntegrationPoint.FOURIER_SINGLE_MODE_SPACED, IntegrationPoint.FOURIER_SINGLE_MODE_SPACED),
        None, (BasisType.FOURIER_SINGLE_MODE, BasisType.FOURIER_SINGLE_MODE, BasisType.FOURIER_SINGLE_MODE)),
])
def test_valid_fourier_single_mode_expansion_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.FourierSingleModeExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For FourierSingleModeExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes",[
    (Elements.TRI,5),
    (Elements.TET,5),
    (Elements.PYR,5),
    (Elements.PRISM,5),
])
def test_invalid_fourier_single_mode_expansion_factory(element:Elements,num_modes:int):
    expansion_factory = nd.FourierSingleModeExpansionFactory()
    
    with pytest.raises(nd.ExpansionValidationException):
        expansion_factory.get_expansion(element,num_modes)

###############################################################################
# Fourier Half Mode Re Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.SEG, 5, None, (5,), (5,), (IntegrationPoint.FOURIER_SINGLE_MODE_SPACED,), None, (BasisType.FOURIER_HALF_MODE_RE,)),
    (Elements.QUAD, 5, None, (5, 5), (5, 5), (IntegrationPoint.FOURIER_SINGLE_MODE_SPACED, IntegrationPoint.FOURIER_SINGLE_MODE_SPACED), None, 
        (BasisType.FOURIER_HALF_MODE_RE, BasisType.FOURIER_HALF_MODE_RE)),
    (Elements.HEX, 5, None, (5, 5, 5), (5, 5, 5), (IntegrationPoint.FOURIER_SINGLE_MODE_SPACED, IntegrationPoint.FOURIER_SINGLE_MODE_SPACED, IntegrationPoint.FOURIER_SINGLE_MODE_SPACED),
        None, (BasisType.FOURIER_HALF_MODE_RE, BasisType.FOURIER_HALF_MODE_RE, BasisType.FOURIER_HALF_MODE_RE)),
])
def test_valid_fourier_half_mode_re_expansion_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.FourierHalfModeReExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For FourierHalfModeReExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes",[
    (Elements.TRI,5),
    (Elements.TET,5),
    (Elements.PYR,5),
    (Elements.PRISM,5),
])
def test_invalid_fourier_half_mode_re_expansion_factory(element:Elements,num_modes:int):
    expansion_factory = nd.FourierHalfModeReExpansionFactory()
    
    with pytest.raises(nd.ExpansionValidationException):
        expansion_factory.get_expansion(element,num_modes)

###############################################################################
# Fourier Half Mode Im Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.SEG, 5, None, (5,), (5,), (IntegrationPoint.FOURIER_SINGLE_MODE_SPACED,), None, (BasisType.FOURIER_HALF_MODE_IM,)),
    (Elements.QUAD, 5, None, (5, 5), (5, 5), (IntegrationPoint.FOURIER_SINGLE_MODE_SPACED, IntegrationPoint.FOURIER_SINGLE_MODE_SPACED), None, 
        (BasisType.FOURIER_HALF_MODE_IM, BasisType.FOURIER_HALF_MODE_IM)),
    (Elements.HEX, 5, None, (5, 5, 5), (5, 5, 5), (IntegrationPoint.FOURIER_SINGLE_MODE_SPACED, IntegrationPoint.FOURIER_SINGLE_MODE_SPACED, IntegrationPoint.FOURIER_SINGLE_MODE_SPACED),
        None, (BasisType.FOURIER_HALF_MODE_IM, BasisType.FOURIER_HALF_MODE_IM, BasisType.FOURIER_HALF_MODE_IM)),
])
def test_valid_fourier_half_mode_im_expansion_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.FourierHalfModeImExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For FourierHalfModeImExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes",[
    (Elements.TRI,5),
    (Elements.TET,5),
    (Elements.PYR,5),
    (Elements.PRISM,5),
])
def test_invalid_fourier_half_mode_im_expansion_factory(element:Elements,num_modes:int):
    expansion_factory = nd.FourierHalfModeImExpansionFactory()
    
    with pytest.raises(nd.ExpansionValidationException):
        expansion_factory.get_expansion(element,num_modes)

###############################################################################
# Chebyshev Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.SEG, 5, None, (5,), (5,), (IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV,), None, (BasisType.CHEBYSHEV,)),
    (Elements.QUAD, 5, None, (5, 5), (5, 5), (IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV, IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV), None, 
        (BasisType.CHEBYSHEV, BasisType.CHEBYSHEV)),
    (Elements.HEX, 5, None, (5, 5, 5), (5, 5, 5), (IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV, IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV, IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV),
        None, (BasisType.CHEBYSHEV, BasisType.CHEBYSHEV, BasisType.CHEBYSHEV)),
])
def test_valid_chebyshev_expansion_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.ChebyshevExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For ChebyshevExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes",[
    (Elements.TRI,5),
    (Elements.TET,5),
    (Elements.PYR,5),
    (Elements.PRISM,5),
])
def test_invalid_chebyshev_expansion_factory(element:Elements,num_modes:int):
    expansion_factory = nd.ChebyshevExpansionFactory()
    
    with pytest.raises(nd.ExpansionValidationException):
        expansion_factory.get_expansion(element,num_modes)

###############################################################################
# Fourier Chebyshev Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.QUAD, 5, None, (5, 5), (5, 5), (IntegrationPoint.FOURIER_EVENLY_SPACED, IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV), None, 
        (BasisType.FOURIER, BasisType.CHEBYSHEV)),
])
def test_valid_fourier_chebyshev_expansion_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.FourierChebyshevExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For FourierChebyshevExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes",[
    (Elements.SEG,5),
    (Elements.TRI,5),
    (Elements.HEX,5),
    (Elements.TET,5),
    (Elements.PYR,5),
    (Elements.PRISM,5),
])
def test_invalid_fourier_chebyshev_expansion_factory(element:Elements,num_modes:int):
    expansion_factory = nd.FourierChebyshevExpansionFactory()
    
    with pytest.raises(nd.ExpansionValidationException):
        expansion_factory.get_expansion(element,num_modes)

###############################################################################
# Chebyshev Fourier Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.QUAD, 5, None, (5, 5), (5, 5), (IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV, IntegrationPoint.FOURIER_EVENLY_SPACED), None, 
        (BasisType.CHEBYSHEV, BasisType.FOURIER)),
])
def test_valid_chebyshev_fourier_expansion_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.ChebyshevFourierExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For ChebyshevFourierExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes",[
    (Elements.SEG,5),
    (Elements.TRI,5),
    (Elements.HEX,5),
    (Elements.TET,5),
    (Elements.PYR,5),
    (Elements.PRISM,5),
])
def test_invalid_chebyshev_fourier_expansion_factory(element:Elements,num_modes:int):
    expansion_factory = nd.ChebyshevFourierExpansionFactory()
    
    with pytest.raises(nd.ExpansionValidationException):
        expansion_factory.get_expansion(element,num_modes)

###############################################################################
# Fourier Modified Factory
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes,fields,expected_modes,expected_points,expected_integration_type,expected_field,expected_basis",[
    (Elements.QUAD, 5, None, (5, 5), (5, 6), (IntegrationPoint.FOURIER_EVENLY_SPACED, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE), None, 
        (BasisType.FOURIER, BasisType.MODIFIED_A)),
])
def test_valid_fourier_modified_expansion_factory(element: Elements, num_modes: int, fields: tuple[str] | None,
                                            expected_modes: tuple[int], expected_points: tuple[int],
                                            expected_integration_type: tuple[IntegrationPoint],
                                            expected_field: tuple[str] | None, expected_basis: tuple[BasisType]):
    expansion_factory = nd.ModifiedFourierExpansionFactory()

    expansion: nd.ExpansionDefinition = expansion_factory.get_expansion(element, num_modes, fields)

    assert_msg_base = f"For ModifiedFourierExpansionFactory factory for element {element}: "

    assert expected_basis == expansion.basis, f"{assert_msg_base} Expected basis {expected_basis}, but got {expansion.basis}"
    assert expected_modes == expansion.num_modes, f"{assert_msg_base} Expected modes {expected_modes}, but got {expansion.num_modes}"
    assert expected_integration_type == expansion.integration_point_type, f"{assert_msg_base} Expected integration type {expected_integration_type}, but got {expansion.integration_point_type}"
    assert expected_points == expansion.num_points, f"{assert_msg_base} Expected points {expected_points}, but got {expansion.num_points}"
    assert expected_field == expansion.fields, f"{assert_msg_base} Expected field {expected_field}, but got {expansion.fields}"

@pytest.mark.unit
@pytest.mark.parametrize("element,num_modes",[
    (Elements.SEG,5),
    (Elements.TRI,5),
    (Elements.HEX,5),
    (Elements.TET,5),
    (Elements.PYR,5),
    (Elements.PRISM,5),
])
def test_invalid_fourier_modified_expansion_factory(element:Elements,num_modes:int):
    expansion_factory = nd.ChebyshevFourierExpansionFactory()
    
    with pytest.raises(nd.ExpansionValidationException):
        expansion_factory.get_expansion(element,num_modes)