import pytest
import yaml
import NekUpload.frontend.config.settings_loader as setting

###############################################################################
# Test storage items methods first
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("name,url,communities",[
    ("US OPSEC","http://add.journalist.com",None),
    ("US OPSEC","http://add.journalist.com",[{"name": "Classified Group", "community_slug": "Very Secure Signal Group"}]),
    ("name","http://url.com",[{"name": "Group1", "community_slug": "1"},{"name": "Group2", "community_slug": "2"}]),
])
def test_db_config_same(name:str,url:str,communities:list[dict[str,str]] | None):
    db1 = setting.DatabaseConfig(name,url,communities)
    db2 = setting.DatabaseConfig(name,url,communities)

    assert id(db1) != id(db2), "These are not two distinct objects, test setup failed"
    assert db1 == db2, "These objects have the same content"

@pytest.mark.unit
def test_db_config_in_list():
    db1 = setting.DatabaseConfig("name1","url1",None)
    db2 = setting.DatabaseConfig("name2","url2",None)
    db3 = setting.DatabaseConfig("name3","url3",[{"name": "name4","community_slug":"5"}])
    db4 = setting.DatabaseConfig("name4","url4",[{"name": "name4","community_slug":"5"},{"name": "name6","community_slug":""}])

    db_list = [db1,db2,db3,db4]
    db1_copy = setting.DatabaseConfig("name1","url1",None)
    db2_copy = setting.DatabaseConfig("name2","url2",None)
    db3_copy = setting.DatabaseConfig("name3","url3",[{"name": "name4","community_slug":"5"}])
    db4_copy = setting.DatabaseConfig("name4","url4",[{"name": "name4","community_slug":"5"},{"name": "name6","community_slug":""}])

    #first check test setup correct
    assert id(db1) != id(db1_copy), "Should be distinct objects with same content"
    assert id(db2) != id(db2_copy), "Should be distinct objects with same content"
    assert id(db3) != id(db3_copy), "Should be distinct objects with same content"
    assert id(db4) != id(db4_copy), "Should be distinct objects with same content"

    assert db1_copy in db_list, "db1_copy should be found in db_list but it is not"
    assert db2_copy in db_list, "db2_copy should be found in db_list but it is not"
    assert db3_copy in db_list, "db3_copy should be found in db_list but it is not"
    assert db4_copy in db_list, "db4_copy should be found in db_list but it is not"

@pytest.mark.unit
@pytest.mark.parametrize("name1,url1,communities1,name2,url2,communities2",[
    ("name","http://url.com",None,"diffname","http://url.com",None),
    ("name","http://url.com",[{"name": "Group", "community_slug": "1"}],"name","http://url.com",None),
])
def test_db_config_diff(name1:str,url1:str,communities1:list[dict[str,str]],
                        name2:str,url2:str,communities2:list[dict[str,str]]):
    
    db1 = setting.DatabaseConfig(name1,url1,communities1)
    db2 = setting.DatabaseConfig(name2,url2,communities2)

    assert db1 != db2, "These objects have the same content"

###############################################################################
# Test config objects are correctly constructed
###############################################################################

@pytest.fixture
def create_config_file(tmp_path):
    filename = tmp_path / "test_config.yaml"

    #do not indent, yaml is space sensitive
    yaml_content = """\
databases:
  - name: AE Datastore
    url: https://data.ae.ic.ac.uk
    communities:
      - name: Nektar++
        community_slug: nektar
"""

    filename.write_text(yaml_content)

    return filename

@pytest.mark.unit
def test_generate_database_config_list():
    yaml_content = {
        "databases": [
            {
                "name": "AE Datastore",
                "url": "https://data.ae.ic.ac.uk",
                "communities": [
                    {
                        "name": "Nektar++",
                        "community_slug": "nektar"
                    }
                ]
            }
        ]
    }

    data: list[setting.DatabaseConfig] = setting._generate_database_config_list(yaml_content)

    expected_data = setting.DatabaseConfig(
                name="AE Datastore",
                url="https://data.ae.ic.ac.uk",
                communities=[
                    {"name": "Nektar++","community_slug": "nektar"},
                    {"name": "Custom", "community_slug":""}, #added by default
                ])
    
    assert expected_data in data, f"Expected configuration not found in loaded data. Expected: {expected_data}, but got: {data}"

@pytest.mark.integration
def test_config_loading(create_config_file):

    data: list[setting.DatabaseConfig] = setting._load_config(create_config_file)

    expected_data = setting.DatabaseConfig(
                name="AE Datastore",
                url="https://data.ae.ic.ac.uk",
                communities=[
                    {"name": "Nektar++","community_slug": "nektar"},
                    {"name": "Custom", "community_slug":""}, #added by default
                ])

    assert expected_data in data, f"Expected configuration not found in loaded data. Expected: {expected_data}, but got: {data}"