from NekUpload.metadata.identifier import Identifier,IdentifierType
import pytest

@pytest.mark.unit
@pytest.mark.parametrize("valid_id, identifier_type",[
    #valid ORCID ids taken from 
    #https://support.orcid.org/hc/en-us/articles/360053289173-Why-does-my-ORCID-iD-have-an-X#:~:text=ORCID%20iD%27s%20that%20end%20with%20an%20X%20are,digits%20of%20your%20identifier%20form%20a%20coherent%20iD.
    ("0000-0001-5109-3700",IdentifierType.ORCID),
    ("0000-0002-1825-0097",IdentifierType.ORCID),
    ("0000-0002-1694-233X",IdentifierType.ORCID),
        
    # Valid ISNI IDs (if adding new types, follow the same format)
    ("000000012146438X", IdentifierType.ISNI),
])
def test_valid(valid_id,identifier_type):
    """Test that valid identifiers are correctly stored."""
    identifier = Identifier(valid_id,identifier_type)

    assert identifier.get_id() == valid_id, f"ID should be {valid_id}, got {identifier.get_id()}"
    assert identifier.get_id_type() == identifier_type, f"ID should be {identifier_type}, got {identifier.get_id_type()}"

@pytest.mark.unit
@pytest.mark.parametrize("description,invalid_id, identifier_type", [
    #Invalid ORCID IDs
    ("too_short_length", "0000-0002-1825-009", IdentifierType.ORCID),
    ("too_long_length", "0000-0002-1825-00971",IdentifierType.ORCID),
    ("invalid_char", "0000-0002-182A-0097",IdentifierType.ORCID),
    ("wrong_checksum", "0000-0002-1825-0098",IdentifierType.ORCID),
    ("missing_hyphens", "0000000218250097",IdentifierType.ORCID),
    ("misplaced_hyphens", "0000000-21825-0097",IdentifierType.ORCID),

    #Invalid ISNI IDs etc.
    ("too_short_length","000000012146438",IdentifierType.ISNI),
    ("too_long_length","000000012146438X1",IdentifierType.ISNI),
    ("hyphenated","0000-0001-2146-438X",IdentifierType.ISNI),
    ("wrong_checksum","0000000121464389",IdentifierType.ISNI),
])
def test_invalid(description,invalid_id,identifier_type):
    """Test that invalid identifiers are correctly rejected"""

    with pytest.raises(ValueError):
        Identifier(invalid_id, identifier_type)
        pytest.fail(f"{description}: {invalid_id} should have raised ValueError")

@pytest.mark.unit
def test_serialisation():
    """Test serialisation creates corerct object"""

    id = "0000-0001-5109-3700"
    identifier = Identifier(id, IdentifierType.ORCID)

    json = identifier.to_json_serialisable()
    json_expected = {
        "id": id,
        "id_type": "orcid"
    }

    assert json == json_expected, f"Expected {json_expected}, got {json}"

@pytest.mark.unit
def test_deserialisation():
    """Test deserialisation creates corret object"""

    json = {
        "id": "0000-0001-5109-3700",
        "id_type": "orcid"
    }

    identifier = Identifier.from_json(json)
    expected_identifier = Identifier("0000-0001-5109-3700",IdentifierType.ORCID)

    assert identifier == expected_identifier, f"Expected {expected_identifier!r}, got "

@pytest.mark.unit
def test_equality_same():
    """Test equality"""

    id = "0000-0001-5109-3700"

    identifier1 = Identifier(id,IdentifierType.ORCID)
    identifier2 = Identifier(id,IdentifierType.ORCID)

    assert identifier1 == identifier2, f"Equality failed, {identifier1!r} is same as {identifier2!r}"

@pytest.mark.unit
def test_equality_diff():
    """Test equality when not same"""

    id1 = "0000-0001-5109-3700"
    id2 = "0000-0002-1825-0097"

    identifier1 = Identifier(id1,IdentifierType.ORCID)
    identifier2 = Identifier(id2,IdentifierType.ORCID)

    assert identifier1 != identifier2, f"Equality failed, {identifier1!r} is differnet from {identifier2!r}"