from typing import Any
import pytest

from NekUpload.metadata.user import InvenioOrgInfo,InvenioUserInfo,InvenioPersonInfo,InvenioUserInfoFactory
from NekUpload.metadata.identifier import Identifier,IdentifierType

"""
Since InvenioUserInfo is a base abstract class
We will test methods indirectly through concrete implementations
"""

###############################################################################
# TEST: InvenioOrgInfo
###############################################################################

@pytest.mark.unit
def test_add_set_name():
    """Test than an organisation name is correctly stored and can be retrieved"""
    org_name = "Imperial College London"
    org_info = InvenioOrgInfo(org_name)
    
    info: dict[str,Any] = org_info.get_info()
    
    assert "name" in info, "Expected 'name' in get_info() output."
    assert info["name"] == org_name, f"Expected name to be '{org_name}', but got '{info['name']}'"

    assert "type" in info, "Expected 'type' key in get_info() output."
    assert info["type"] == "organizational", f"Expected type to be 'organizational', but got '{info['type']}'"

@pytest.mark.unit
def test_add_one_identifier():
    """Test that adding one identifier from Identifier class is successful, and can be retrieved successfully"""
    org_name = "Imperial College London"
    org_info = InvenioOrgInfo(org_name)
    
    #create an identifier
    orcid = "0000-0002-1694-233X"
    identifier = Identifier(orcid,IdentifierType.ORCID)

    org_info.add_identifier(identifier)

    info: dict[str,Any] = org_info.get_info()
    
    assert "identifiers" in info, "Expected 'identifiers' in get_info() output."

    stored_identifiers: list[str] = info["identifiers"]
    
    assert isinstance(stored_identifiers, list), "Expected 'identifiers' to be a list."
    assert len(stored_identifiers) == 1, f"Expected 1 identifier, got {len(stored_identifiers)} identifiers."

    assert "scheme" in stored_identifiers[0], "Expected 'scheme' to be present in the identifier information"
    assert stored_identifiers[0]["scheme"] == "orcid", f"Expected identifier scheme orcid, got {stored_identifiers[0]['scheme']}"
    assert "identifier" in stored_identifiers[0], "Expected 'identifier' to be present in the identifier information"
    assert stored_identifiers[0]["identifier"] == orcid, f"Expected identifier {orcid}, got {stored_identifiers[0]['identifier']}"

@pytest.mark.unit
def test_add_duplicate_identifiers():
    """Test that adding duplicate identifiers leads to successful rejection (i.e. error should be thrown)"""
    org_name = "Imperial College London"
    org_info = InvenioOrgInfo(org_name)
    
    #create a identifiers
    orcid = "0000-0002-1694-233X"
    identifier1 = Identifier(orcid,IdentifierType.ORCID)
    identifier2 = Identifier(orcid,IdentifierType.ORCID)

    org_info.add_identifier(identifier1)

    with pytest.raises(ValueError):
        org_info.add_identifier(identifier2)

@pytest.mark.unit
def test_serialisation():
    """Test that object is successfully serialised with all relevant information"""
    org_name = "Imperial College London"
    org_info = InvenioOrgInfo(org_name)
    
    #create an identifier
    orcid = "0000-0002-1694-233X"
    identifier = Identifier(orcid,IdentifierType.ORCID)

    org_info.add_identifier(identifier)

    json: dict[str,Any] = org_info.to_json_serialisable()

    json_expected = {
        "type": "organizational",
        "name": org_name,
        "identifiers": [identifier.to_json_serialisable()]
    }

    assert json == json_expected, f"Expected {json_expected}, but got {json}"
    
@pytest.mark.unit
def test_deserialisation():
    """Test that the object is successfully deserialised and constructs full object"""
    #create an identifier
    orcid = "0000-0002-1694-233X"
    identifier = Identifier(orcid,IdentifierType.ORCID)

    json = {
        "type": "organizational",
        "name": "Imperial College London",
        "identifiers": [identifier.to_json_serialisable()]
    }

    org = InvenioOrgInfo.from_json(json)

    expected_object = InvenioOrgInfo("Imperial College London")
    expected_object.add_identifier(identifier)

    assert org == expected_object, f"Expected {expected_object}, got {org}"

@pytest.mark.unit
def test_same():
    """Test that equality operation works correctly"""
    name = "UCL"

    org1 = InvenioOrgInfo(name)
    org2 = InvenioOrgInfo(name)

    assert org1 == org2

    orcid = "0000-0002-1694-233X"
    identifier = Identifier(orcid,IdentifierType.ORCID)

    org1.add_identifier(identifier)
    org2.add_identifier(identifier)

    assert org1 == org2

@pytest.mark.unit
def test_diff_case1():
    """Test that NOT equality works correctly, here different names"""
    name1 = "UCL"
    name2 = "KCL"

    org1 = InvenioOrgInfo(name1)
    org2 = InvenioOrgInfo(name2)

    assert org1 != org2

@pytest.mark.unit
def test_diff_case2():
    """Test that NOT equality works correctly, here different identifiers"""
    name = "UCL"

    org1 = InvenioOrgInfo(name)
    org2 = InvenioOrgInfo(name)

    id1 = "0000-0001-5109-3700"
    id2 = "0000-0002-1825-0097"

    org1.add_identifier(Identifier(id1,IdentifierType.ORCID))
    org2.add_identifier(Identifier(id2,IdentifierType.ORCID))

    assert org1 != org2

###############################################################################
# TEST: InvenioPersonInfo
###############################################################################

@pytest.mark.unit
def test_constructor():
    """Test whether default, just adding name, works"""
    given_name = "John"
    family_name = "Smith"

    user_info = InvenioPersonInfo(given_name,family_name)

    info: dict[str,Any] = user_info.get_info()

    assert "given_name" in info, "Expected 'given_name' in info"
    assert info["given_name"] == given_name, f"Expected {given_name}, got {info['given_name']}"

    assert "family_name" in info, "Expected 'family_name' in info"
    assert info["family_name"] == family_name, f"Expected {family_name}, got {info['family_name']}"

    assert "type" in info, "Expected 'type' in info"
    assert info["type"] == "personal", f"Expected 'personal', got {info['type']}"

@pytest.mark.unit
def test_add_one_identifier():
    """Test whether adding a single identifier works correctly"""
    given_name = "John"
    family_name = "Smith"

    user_info = InvenioPersonInfo(given_name,family_name)
    
    #create an identifier
    orcid = "0000-0002-1694-233X"
    identifier = Identifier(orcid,IdentifierType.ORCID)

    user_info.add_identifier(identifier)

    info: dict[str,Any] = user_info.get_info()
    
    assert "identifiers" in info, "Expected 'identifiers' in get_info() output."

    stored_identifiers: list[str] = info["identifiers"]
    
    assert isinstance(stored_identifiers, list), "Expected 'identifiers' to be a list."
    assert len(stored_identifiers) == 1, f"Expected 1 identifier, got {len(stored_identifiers)} identifiers."

    assert "scheme" in stored_identifiers[0], "Expected 'scheme' to be present in the identifier information"
    assert stored_identifiers[0]["scheme"] == "orcid", f"Expected identifier scheme orcid, got {stored_identifiers[0]['scheme']}"
    assert "identifier" in stored_identifiers[0], "Expected 'identifier' to be present in the identifier information"
    assert stored_identifiers[0]["identifier"] == orcid, f"Expected identifier {orcid}, got {stored_identifiers[0]['identifier']}"

@pytest.mark.unit
def test_add_duplicate_identifiers():
    """Test whether succesful error raised on entering duplicate identifiers"""
    given_name = "John"
    family_name = "Smith"

    user_info = InvenioPersonInfo(given_name,family_name)
    
    #create a identifiers
    orcid = "0000-0002-1694-233X"
    identifier1 = Identifier(orcid,IdentifierType.ORCID)
    identifier2 = Identifier(orcid,IdentifierType.ORCID)

    user_info.add_identifier(identifier1)

    with pytest.raises(ValueError):
        user_info.add_identifier(identifier2)

@pytest.mark.unit
def test_serialisation():
    """Test whether object successfully serialised"""
    given_name = "John"
    family_name = "Smith"

    user_info = InvenioPersonInfo(given_name,family_name)
    
    #create an identifier
    orcid = "0000-0002-1694-233X"
    identifier = Identifier(orcid,IdentifierType.ORCID)

    user_info.add_identifier(identifier)

    json: dict[str,Any] = user_info.to_json_serialisable()

    json_expected = {
        "type": "personal",
        "given_name": given_name,
        "family_name": family_name,
        "identifiers": [identifier.to_json_serialisable()]
    }

    assert json == json_expected, f"Expected {json}, got {json_expected}"

@pytest.mark.unit
def test_deserialisation():    
    """Test whether object succesfully deserialised"""
    #create an identifier
    orcid = "0000-0002-1694-233X"
    identifier = Identifier(orcid,IdentifierType.ORCID)

    json = {
        "type": "personal",
        "given_name": "John",
        "family_name": "Smith",
        "identifiers": [identifier.to_json_serialisable()]
    }

    person = InvenioPersonInfo.from_json(json)

    expected_object = InvenioPersonInfo("John","Smith")
    expected_object.add_identifier(identifier)

    assert person == expected_object, f"Expected {expected_object}, got {person}"

@pytest.mark.unit
def test_same_only_name():
    """Test whether equality works"""
    given_name = "Winston"
    last_name = "Churchill"

    person1 = InvenioPersonInfo(given_name,last_name)
    person2 = InvenioPersonInfo(given_name,last_name)

    assert person1 == person2, f"For only name, expected to be equal"
    orcid = "0000-0002-1694-233X"
    identifier = Identifier(orcid,IdentifierType.ORCID)

    person1.add_identifier(identifier)
    person2.add_identifier(identifier)

    assert person1 == person2

@pytest.mark.unit
def test_same_name_and_id():
    """Test whether equality works, including ids"""
    given_name = "Winston"
    last_name = "Churchill"
    orcid = "0000-0002-1694-233X"
    identifier = Identifier(orcid,IdentifierType.ORCID)

    person1 = InvenioPersonInfo(given_name,last_name)
    person2 = InvenioPersonInfo(given_name,last_name)
    person1.add_identifier(identifier)
    person2.add_identifier(identifier)

    assert person1 == person2, "For name and ids, expected to be equal"

@pytest.mark.unit
def test_diff_case1():
    """Test whether NOT equality works, for diff names"""
    given_name1 = "Josef"
    last_name1 = "Stalin"
    given_name2 = "Adolf"
    last_name2 = "Hitler"

    person1 = InvenioPersonInfo(given_name1,last_name1)
    person2 = InvenioPersonInfo(given_name2,last_name2)

    assert person1 != person2, "Adolf Hitler is not same person as Adolf Hitler (although close)"

@pytest.mark.unit
def test_diff_case2():
    """Test whether NOT equality works, for diff ids"""
    given_name = "Josef"
    last_name = "Stalin"

    person1 = InvenioPersonInfo(given_name,last_name)
    person2 = InvenioPersonInfo(given_name,last_name)

    id1 = "0000-0001-5109-3700"
    id2 = "0000-0002-1825-0097"

    person1.add_identifier(Identifier(id1,IdentifierType.ORCID))
    person2.add_identifier(Identifier(id2,IdentifierType.ORCID))

    assert person1 != person2, "Ids should be different"

###############################################################################
# TEST DESERIALISATION FACTORY
###############################################################################

@pytest.mark.unit
def test_deserialisation_factory_org():    
    orcid = "0000-0002-1694-233X"
    identifier = Identifier(orcid,IdentifierType.ORCID)

    json = {
        "type": "organizational",
        "name": "Imperial College London",
        "identifiers": [identifier.to_json_serialisable()]
    }

    user_info: (InvenioOrgInfo | InvenioPersonInfo) = InvenioUserInfoFactory.create_from_json(json)

    assert isinstance(user_info,InvenioOrgInfo)

@pytest.mark.unit
def test_deserialisation_factory_person():
    #create an identifier
    orcid = "0000-0002-1694-233X"
    identifier = Identifier(orcid,IdentifierType.ORCID)

    json = {
        "type": "personal",
        "given_name": "John",
        "family_name": "Smith",
        "identifiers": [identifier.to_json_serialisable()]
    }

    user_info: (InvenioOrgInfo | InvenioPersonInfo) = InvenioUserInfoFactory.create_from_json(json)

    assert isinstance(user_info,InvenioPersonInfo)