import pytest
import requests
from collections import namedtuple
from dotenv import load_dotenv
import os
import logging
from typing import Any
from NekUpload.upload import InvenioRDM
import NekUpload.upload.invenio_rdm_api as InvenioAPI
from NekUpload.metadata import *

###############################################################################
# These all require internet connection to a test instance of database
# So skip tests if database targets or environment variables not present or
# no internet
###############################################################################
load_dotenv()

def is_repository_available(url:str) -> bool:
    try:
        response = requests.get(url,timeout=5)
        return response.status_code == 200
    except requests.RequestException:
        return False

def are_required_env_vars_set(*env_vars) -> bool:
    return all(var in os.environ for var in env_vars)

#db targets and associated variables
DBInfo = namedtuple("DBInfo",["name","url","token_env_var_name","community_slug_env_var_name"])

INVENIO_RDM_DEMO_V12 = DBInfo(
    name="InvenioRDM Demo V12",
    url="https://inveniordm.web.cern.ch",
    token_env_var_name="INVENIO_RDM_DEMO_TOKEN",
    community_slug_env_var_name="INVENIO_RDM_TEST_COMMUNITY_SLUG"
) 

AE_DATASTORE_DEV = DBInfo(
    name="AEDatastore Dev",
    url="https://data-dev.ae.ic.ac.uk",
    token_env_var_name="AE_DATASTORE_DEV_TOKEN",
    community_slug_env_var_name="AE_DATASTORE_DEV_COMMUNITY_SLUG"
)

IS_TEST_INVENIO_RDM: bool = (is_repository_available(INVENIO_RDM_DEMO_V12.url) and
                            are_required_env_vars_set(INVENIO_RDM_DEMO_V12.token_env_var_name,INVENIO_RDM_DEMO_V12.community_slug_env_var_name))
IS_TEST_AE_DATASTORE: bool = (is_repository_available(AE_DATASTORE_DEV.url) and
                            are_required_env_vars_set(AE_DATASTORE_DEV.token_env_var_name,AE_DATASTORE_DEV.community_slug_env_var_name))

EnvVarVals = namedtuple("EnvVarVals",["token","community_slug"])

def get_env_helper(db_info: DBInfo) -> EnvVarVals:
    token = os.getenv(db_info.token_env_var_name)
    community_slug = os.getenv(db_info.community_slug_env_var_name)
    return EnvVarVals(token,community_slug)

@pytest.fixture
def setup_file(tmp_path):
    filename = tmp_path / "file.txt"

    filename.write_text("This is a test file from NekUpload. If you see this, delete the record.")

    return str(filename)

def cleanup_record_draft(url:str,token:str,record_id: str):
    """Custom function to delete the draft."""
    try:
        InvenioAPI.delete_draft(url, token, record_id)
        logging.info(f"Draft record {record_id} deleted successfully.")
    except Exception as e:
        logging.error(f"Failed to delete draft record {record_id}: {e}")

def cleanup_community_submission(url:str,token:str,record_id:str,request_id:str):
    try:
        InvenioAPI.cancel_review_request(url,token,request_id)
        InvenioAPI.delete_review_request(url,token,record_id)
        logging.info(f"Draft record {record_id} review request deleted successfully.")
    except Exception as e:
        logging.error(f"Failed to delete review request for {record_id}: {e}")


###############################################################################
# Tests start here
###############################################################################

@pytest.mark.integration
@pytest.mark.internet
@pytest.mark.parametrize("db_info",[
        pytest.param(
            INVENIO_RDM_DEMO_V12,
            marks=pytest.mark.skipif(not IS_TEST_INVENIO_RDM, reason="InvenioRDM demo not accessible"),
        ),    
        pytest.param(
            AE_DATASTORE_DEV,
            marks=pytest.mark.skipif(not IS_TEST_AE_DATASTORE, reason="AE_DATASTORE not accessible"),
        ),        
])
def test_invenio_rdm_upload_check_metadata(request,setup_file,db_info:DBInfo):
    #set up metadata
    title = "THIS IS A TEST: DELETE ON SIGHT"
    creator = InvenioPersonInfo("John","Doe")
    version = "1.0.0"
    publisher = "NekUpload"
    description = "DELETE ME NOW!!!"
    publication_date = "2020-01-01"

    metadata = InvenioMetadata(title,publication_date,[creator],ResourceType.DATASET)
    metadata.add_version(version)
    metadata.add_publisher(publisher)
    metadata.add_description(description)

    metadata_payload = metadata.get_metadata_payload()

    db = InvenioRDM()
    token,_ = get_env_helper(db_info)
    url = db_info.url

    try:
        db.create_record(url,token,[setup_file],metadata_payload)
    except Exception:
        #failure during get_record, if got to point where draft created, delete
        #otherwise do nothing
        if record_id := db.record_id:
            cleanup_record_draft(url,token,record_id)
        raise
    else:
        record_id = db.record_id
        #this will clean up the draft
        def finalizer():
            cleanup_record_draft(url, token, record_id)

        request.addfinalizer(finalizer)

        # Now grab the record metadata and check information
        response = InvenioAPI.get_draft_record(url, token, record_id)
        data = response.json()

        # Check metadata is correctly uploaded
        metadata_json = data["metadata"]
        record_title = metadata_json.get("title")
        record_publisher = metadata_json.get("publisher")
        record_publication_date = metadata_json.get("publication_date")
        record_version = metadata_json.get("version")
        record_description = metadata_json.get("description")

        assert record_title == title, f"Expected title '{title}', but got '{record_title}'"
        assert record_publication_date == publication_date, f"Expected publication date '{publication_date}', but got '{record_publication_date}'"
        assert record_publisher == publisher, f"Expected publisher '{publisher}', but got '{record_publisher}'"
        assert record_version == version, f"Expected version '{version}', but got '{record_version}'"
        assert record_description == description, f"Expected description '{description}', but got '{record_description}'"

        creator_info = metadata_json["creators"][0]["person_or_org"]
        assert creator_info["given_name"] == "John", f"Expected given name 'John', but got '{creator_info['given_name']}'"
        assert creator_info["family_name"] == "Doe", f"Expected family name 'Doe', but got '{creator_info['family_name']}'"
        assert metadata_json["resource_type"]["id"] == ResourceType.DATASET.value, (
            f"Expected resource type '{ResourceType.DATASET.value}', but got '{metadata_json['resource_type']['id']}'"
        )

@pytest.fixture
def metadata() -> InvenioMetadata:
    title = "THIS IS A TEST: DELETE ON SIGHT"
    creator = InvenioPersonInfo("John","Doe")
    version = "1.0.0"
    publisher = "NekUpload"
    description = "DELETE ME NOW!!!"
    publication_date = "2020-01-01"

    metadata = InvenioMetadata(title,publication_date,[creator],ResourceType.DATASET)
    metadata.add_version(version)
    metadata.add_publisher(publisher)
    metadata.add_description(description)

    return metadata

@pytest.mark.integration
@pytest.mark.internet
@pytest.mark.parametrize("db_info",[
        pytest.param(
            INVENIO_RDM_DEMO_V12,
            marks=pytest.mark.skipif(not IS_TEST_INVENIO_RDM, reason="InvenioRDM demo not accessible"),
        ),    
        pytest.param(
            AE_DATASTORE_DEV,
            marks=pytest.mark.skipif(not IS_TEST_AE_DATASTORE, reason="AE_DATASTORE not accessible"),
        ),        
])
def test_invenio_rdm_submit_to_community(setup_file,metadata:InvenioMetadata,db_info:DBInfo):
    metadata_payload = metadata.get_metadata_payload()

    db = InvenioRDM()
    token,community_slug = get_env_helper(db_info)
    url = db_info.url

    #set up the record
    try:
        db.create_record(url,token,[setup_file],metadata_payload)
    except Exception as e:
        #failure during get_record, if got to point where draft created, delete
        #otherwise do nothing
        if record_id := db.record_id:
            cleanup_record_draft(url,token,record_id)
        pytest.fail("Test setup failed")

    #now perform test
    try:
        db.submit_to_community(url,token,community_slug)
    except:
        raise
    finally:
        if db.record_id and db.request_id:
            cleanup_community_submission(url,token,db.record_id,db.request_id)
            cleanup_record_draft(url,token,db.record_id)