import pytest
import requests
from typing import Any
import NekUpload.upload.invenio_rdm_api as InvenioAPI

###############################################################################
# Testing philosophy is that we need to ensure the rules for the API wrapper
# are correct
# This means ensuring the following:
# - Check API request (url,header,payload) is good
# - Check correct response code for good and bad paths
# This does not involve checking content of the request, but important content
# should be included in the response mocking to provide others with understanding
# of what parts of the response object to look out for
###############################################################################

###############################################################################
# Fixtures for common test items
###############################################################################

@pytest.fixture
def sample_metadata() -> dict[str,Any]:
    metadata =  {
                "creators": [
                {
                    "person_or_org": {
                    "family_name": "Brown",
                    "given_name": "Troy",
                    "type": "personal"
                    }
                },
                {
                    "person_or_org": {
                    "family_name": "Collins",
                    "given_name": "Thomas",
                    "identifiers": [
                        {"scheme": "orcid", "identifier": "0000-0002-1825-0097"}
                    ],
                    "name": "Collins, Thomas",
                    "type": "personal"
                    },
                    "affiliations": [
                    {
                        "id": "01ggx4157",
                        "name": "Entity One"
                    }
                    ]
                }
                ],
            }
    
    return metadata

@pytest.fixture
def sample_record_id() -> str:
    return "id12345"

@pytest.fixture
def sample_host_name() -> str:
    return "https://api.example.com"

@pytest.fixture
def sample_token() -> str:
    return "token12345"

@pytest.fixture
def sample_community_slug() -> str:
    return "slug"

@pytest.fixture
def sample_community_uuid() -> str:
    return "uuid"

@pytest.fixture
def sample_request_id() -> str:
    return "request_id_1234"

###############################################################################
# Test functions first before API call
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("status_code,exception",[
    (400,InvenioAPI.ClientError),
    (401,InvenioAPI.ClientError),
    (500,InvenioAPI.ServerError),
    (501,InvenioAPI.ServerError),
    (300,InvenioAPI.APIError),
])
def test_raise_status_error(mocker,status_code: int,exception: Exception):#
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = status_code

    with pytest.raises(exception):
        InvenioAPI.raise_status_error(mock_response)
        pytest.fail(f"Status code {status_code} should have raise {exception}")

###############################################################################
# Create Draft Record API
###############################################################################

@pytest.fixture
def create_draft_record_mock_response(mocker,sample_metadata,sample_host_name,sample_record_id) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 201
    mock_response.json.return_value = {
        "metadata": sample_metadata,
        "id": sample_record_id,
        "links": {
            "latest": f"{sample_host_name}/api/records/{sample_record_id}/versions/latest",
        }
    }

    mock_post = mocker.patch("requests.post",return_value = mock_response)

    return mock_post,mock_response

@pytest.mark.unit
def test_valid_create_draft_record(mocker,create_draft_record_mock_response,
                                sample_metadata,sample_host_name,sample_token):
    mock_post,mock_response = create_draft_record_mock_response

    response = InvenioAPI.create_draft_record(
        url=sample_host_name,
        token=sample_token,
        metadata=sample_metadata
    )

    assert response.status_code == 201, f"Create Draft Record API failed to raise 201"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_post.assert_called_once()

    #now verify call parameters
    call_args = mock_post.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + "/api/records"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {sample_token}"
    }
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

    #check certain key parameters of the body
    called_body: dict[str,Any] = call_args[1]["json"]
    assert called_body.get("metadata",None) == sample_metadata, f"Expected metadata to be 'metadata: {sample_metadata}', got {called_body}"
    assert called_body.get("files",None) == {"enabled": True}, f"Expected files_enabled to be True, got {called_body}"

@pytest.mark.unit
def test_valid_create_draft_record_file_unenabled(mocker,create_draft_record_mock_response,
                                sample_metadata,sample_host_name,sample_token):
    mock_post,mock_response = create_draft_record_mock_response

    response = InvenioAPI.create_draft_record(
        url=sample_host_name,
        token=sample_token,
        metadata=sample_metadata,
        upload_file_enabled=False
    )

    assert response.status_code == 201, f"Create Draft Record API failed to raise 201"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_post.assert_called_once()

    #now verify call parameters
    call_args = mock_post.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + "/api/records"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {sample_token}"
    }
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

    called_body: dict[str,Any] = call_args[1]["json"]
    assert called_body.get("metadata",None) == sample_metadata, f"Expected metadata to be 'metadata: {sample_metadata}', got {called_body}"
    assert called_body.get("files",None) == {"enabled": False}, f"Expected files_enabled to be True, got {called_body}"

###############################################################################
# Update Draft Record API
###############################################################################

@pytest.fixture
def update_draft_record_mock_response(mocker,sample_metadata,sample_host_name,sample_record_id) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 200
    mock_response.json.return_value = {
        "metadata": sample_metadata,
        "id": sample_record_id,
        "links": {
            "latest": f"{sample_host_name}/api/records/{sample_record_id}/versions/latest",
        }
    }

    mock_put = mocker.patch("requests.put",return_value = mock_response)

    return mock_put,mock_response

@pytest.mark.unit
def test_valid_update_draft_record(mocker,update_draft_record_mock_response,
                                sample_metadata,sample_host_name,sample_token,
                                sample_record_id):
    mock_put,mock_response = update_draft_record_mock_response

    response = InvenioAPI.update_draft_record(
        url=sample_host_name,
        token=sample_token,
        record_id=sample_record_id,
        metadata=sample_metadata,
    )

    assert response.status_code == 200, f"Update Draft Record API failed to raise 200"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_put.assert_called_once()

    #now verify call parameters
    call_args = mock_put.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/records/{sample_record_id}/draft"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {sample_token}"
    }
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

    #check certain key parameters of the body
    called_body: dict[str,Any] = call_args[1]["json"]
    assert called_body.get("metadata",None) == sample_metadata, f"Expected metadata to be 'metadata: {sample_metadata}', got {called_body}"
    assert called_body.get("files",None) == {"enabled": True}, f"Expected files_enabled to be True, got {called_body}"

###############################################################################
# Create New Record Version API
###############################################################################

@pytest.fixture
def create_new_record_version_mock_response(mocker,sample_metadata,sample_host_name,sample_record_id) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 201
    mock_response.json.return_value = {
        "metadata": sample_metadata,
        "id": sample_record_id,
        "links": {
            "latest": f"{sample_host_name}/api/records/{sample_record_id}/versions/latest",
        }
    }

    mock_post = mocker.patch("requests.post",return_value = mock_response)

    return mock_post,mock_response

@pytest.mark.unit
def test_new_record_version_record(mocker,create_new_record_version_mock_response,
                                sample_host_name,sample_token,sample_record_id):
    mock_post,mock_response = create_new_record_version_mock_response

    response = InvenioAPI.create_new_record_version(sample_host_name,sample_token,sample_record_id)

    assert response.status_code == 201, f"Update Draft Record API failed to raise 201"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_post.assert_called_once()

    #now verify call parameters
    call_args = mock_post.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/records/{sample_record_id}/versions"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {sample_token}"
    }
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

###############################################################################
# Reserve DOI Draft API
###############################################################################

@pytest.fixture
def create_reserve_doi_draft_mock_response(mocker,sample_metadata,sample_host_name,sample_record_id) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 201
    mock_response.json.return_value = {
        "metadata": sample_metadata,
        "id": sample_record_id,
        "links": {
            "latest": f"{sample_host_name}/api/records/{sample_record_id}/versions/latest",
        },
        "pids": {
            "doi": {
                "identifier": "something",
                "provider": "somewhere",
                "client": "someone"
            }
        }
    }

    mock_post = mocker.patch("requests.post",return_value = mock_response)

    return mock_post,mock_response

@pytest.mark.unit
def test_reserve_doi_draft_record(mocker,create_reserve_doi_draft_mock_response,
                                sample_host_name,sample_token,sample_record_id):
    mock_post,mock_response = create_reserve_doi_draft_mock_response

    response = InvenioAPI.reserve_doi_draft(sample_host_name,sample_token,sample_record_id)

    assert response.status_code == 201, f"Update Draft Record API failed to raise 201"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_post.assert_called_once()

    #now verify call parameters
    call_args = mock_post.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/records/{sample_record_id}/draft/pids/doi"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {sample_token}"
    }
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

###############################################################################
# Prepare File Upload API
###############################################################################

@pytest.fixture
def create_prepare_file_upload_mock_response(mocker,sample_metadata,sample_host_name,sample_record_id) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 201
    mock_response.json.return_value = {
        "metadata": sample_metadata,
        "id": sample_record_id,
        "links": {
            "latest": f"{sample_host_name}/api/records/{sample_record_id}/versions/latest",
        },
        #just an example, obviosuly depends on what files are prepared for upload
        "entries":[
            {
                "key": "file1"
            },
            {
                "key": "file2"
            }
        ]
    }

    mock_post = mocker.patch("requests.post",return_value = mock_response)

    return mock_post,mock_response

@pytest.mark.unit
@pytest.mark.parametrize("filename_list,expected_body",[
    (["file1.jpg"],[{"key": "file1.jpg"}]),
    (["file1.jpg","file2.png"],[{"key": "file1.jpg"},{"key": "file2.png"}]),
])
def test_prepare_file_upload_draft_record(mocker,create_prepare_file_upload_mock_response,
                                sample_host_name,sample_token,sample_record_id,
                                filename_list:list[str],expected_body:list[dict[str,str]]):
    mock_post,mock_response = create_prepare_file_upload_mock_response

    #quirk of the call:
    response = InvenioAPI.prepare_file_upload(sample_host_name,sample_token,
                                            sample_record_id,filename_list)

    assert response.status_code == 201, f"Update Draft Record API failed to raise 201"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_post.assert_called_once()

    #now verify call parameters
    call_args = mock_post.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/records/{sample_record_id}/draft/files"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {sample_token}"
    }
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

    #check certain key parameters of the body
    called_body: dict[str,Any] = call_args[1]["json"]
    assert called_body == expected_body, f"Expected body to be {expected_body}, got {called_body}"

###############################################################################
# Upload File API
###############################################################################

@pytest.fixture
def create_upload_file_mock_response(mocker,sample_metadata,sample_host_name,sample_record_id) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 200
    mock_response.json.return_value = {
        #depends on what file is being targeted for uplaod
        "key": "file1",
        "content": f"{sample_host_name}/api/records/{sample_record_id}/draft/files/file1/content",
        "self": f"{sample_host_name}/api/records/{sample_record_id}/draft/files/file1",
        "content": f"{sample_host_name}/api/records/{sample_record_id}/draft/files/file1/commit"
    }

    mock_put = mocker.patch("requests.put",return_value = mock_response)

    return mock_put,mock_response

@pytest.mark.unit
@pytest.mark.parametrize("filepath,filename",[
    ("file1.jpg","file1.jpg"),
    ("path/to/file2.jpg","file2.jpg"),
])
def test_valid_upload_file_record(mocker,tmp_path,create_upload_file_mock_response,
                                sample_host_name,sample_token,sample_record_id,
                                filepath:str,filename:str):
    mock_put,mock_response = create_upload_file_mock_response

    #create temp file for upload
    full_file_path = tmp_path / filepath
    full_file_path.parent.mkdir(parents=True, exist_ok=True)  # Creates the directory structure
    full_file_path.write_text("This is a temporary test file. Will cleaned up by pytest")
    full_file_path = str(full_file_path) #function expects string, not Path

    response = InvenioAPI.upload_file(sample_host_name,sample_token,sample_record_id,full_file_path)

    assert response.status_code == 200, f"Update Draft Record API failed to raise 200"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_put.assert_called_once()

    #now verify call parameters
    call_args = mock_put.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/records/{sample_record_id}/draft/files/{filename}/content"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers = {
        "Content-Type": "application/octet-stream",
        "Authorization": f"Bearer {sample_token}"
    }
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

###############################################################################
# Commit File API
###############################################################################

@pytest.fixture
def create_commit_file_mock_response(mocker,sample_metadata,sample_host_name,sample_record_id) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 200
    mock_response.json.return_value = {
        #depends on what file is being targeted for uplaod
        "key": "file1",
        "content": f"{sample_host_name}/api/records/{sample_record_id}/draft/files/file1/content",
        "self": f"{sample_host_name}/api/records/{sample_record_id}/draft/files/file1",
        "content": f"{sample_host_name}/api/records/{sample_record_id}/draft/files/file1/commit",
        "created": "2020-11-27 11:17:10.998919",
        "file_id": "2151fa94-6dc3-4965-8df9-ec73ceb9175c",
        "version_id": "57ad8c66-b934-49c9-a46f-38bf5aa0374f",
        "bucket_id": "90b5b318-114a-4b87-bc9d-0d018b9363d3",
    }

    mock_post = mocker.patch("requests.post",return_value = mock_response)

    return mock_post,mock_response

@pytest.mark.unit
def test_commit_file_upload_record(mocker,create_commit_file_mock_response,
                                sample_host_name,sample_token,sample_record_id):
    mock_post,mock_response = create_commit_file_mock_response

    test_file = "file1.jpg"
    response = InvenioAPI.commit_file_upload(sample_host_name,sample_token,
                                            sample_record_id,test_file)

    assert response.status_code == 200, f"Update Draft Record API failed to raise 200"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_post.assert_called_once()

    #now verify call parameters
    call_args = mock_post.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/records/{sample_record_id}/draft/files/{test_file}/commit"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers = {"Authorization": f"Bearer {sample_token}"}
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

###############################################################################
# Publish Draft API
###############################################################################

@pytest.fixture
def create_publish_draft_mock_response(mocker,sample_metadata,sample_host_name,sample_record_id) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 202
    mock_response.json.return_value = {
        "metadata": sample_metadata,
        "id": sample_record_id,
        "links": {
            "latest": f"{sample_host_name}/api/records/{sample_record_id}/versions/latest",
        },
        #just an example, obviosuly depends on what files are prepared for upload
        "entries":[
            {
                "key": "file1"
            },
            {
                "key": "file2"
            }
        ]
    }

    mock_post = mocker.patch("requests.post",return_value = mock_response)

    return mock_post,mock_response

@pytest.mark.unit
def test_commit_file_upload_record(mocker,create_publish_draft_mock_response,
                                sample_host_name,sample_token,sample_record_id):
    mock_post,mock_response = create_publish_draft_mock_response

    response = InvenioAPI.publish_draft(sample_host_name,sample_token,sample_record_id)

    assert response.status_code == 202, f"Update Draft Record API failed to raise 202"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_post.assert_called_once()

    #now verify call parameters
    call_args = mock_post.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/records/{sample_record_id}/draft/actions/publish"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers = {"Authorization": f"Bearer {sample_token}"}
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

###############################################################################
# Delete Draft API
###############################################################################

@pytest.fixture
def create_delete_draft_mock_response(mocker) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 204
    mock_response.json.return_value = {}

    mock_delete = mocker.patch("requests.delete",return_value = mock_response)

    return mock_delete,mock_response

@pytest.mark.unit
def test_delete_draft_record(mocker,create_delete_draft_mock_response,
                                sample_host_name,sample_token,sample_record_id):
    mock_delete,mock_response = create_delete_draft_mock_response

    response = InvenioAPI.delete_draft(sample_host_name,sample_token,sample_record_id)

    assert response.status_code == 204, f"Update Draft Record API failed to raise 204"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_delete.assert_called_once()

    #now verify call parameters
    call_args = mock_delete.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/records/{sample_record_id}/draft"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers = {"Authorization": f"Bearer {sample_token}"}
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

###############################################################################
# Get Community API
###############################################################################

@pytest.fixture
def create_get_community_mock_response(mocker,sample_host_name,sample_community_slug,sample_community_uuid) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 200
    mock_response.json.return_value = {
        "slug": sample_community_slug,
        "uuid": sample_community_uuid,
        "links": {
            "self": f"{sample_host_name}/api/communities/{sample_community_uuid}"
        }
    }

    mock_get = mocker.patch("requests.get",return_value = mock_response)

    return mock_get,mock_response

@pytest.mark.unit
def test_get_community(mocker,create_get_community_mock_response,
                    sample_host_name,sample_token,sample_community_slug):
    mock_get,mock_response = create_get_community_mock_response

    response = InvenioAPI.get_community(sample_host_name,sample_token,sample_community_slug)

    assert response.status_code == 200, f"Update Draft Record API failed to raise 202"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_get.assert_called_once()

    #now verify call parameters
    call_args = mock_get.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/communities/{sample_community_slug}"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers = {"Authorization": f"Bearer {sample_token}"}
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

###############################################################################
# Submit Record To Community API
###############################################################################

@pytest.fixture
def create_submit_record_to_community_mock_response(mocker,sample_host_name,sample_record_id) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 200
    request_id = "1234"
    mock_response.json.return_value = {
        "id": request_id,
        "links": {
            "actions": {
                "submit": f"{sample_host_name}/api/requests/{request_id}/actions/submit"
            },
            "self": f"{sample_host_name}/api/requests/{request_id}"
        },
        "type": "community-submission",
        "topic": {
            "record": sample_record_id
        }
    }

    mock_put = mocker.patch("requests.put",return_value = mock_response)

    return mock_put,mock_response

@pytest.mark.unit
def test_submit_record_to_community(mocker,create_submit_record_to_community_mock_response,
                    sample_host_name,sample_token,sample_community_uuid,sample_record_id):
    mock_put,mock_response = create_submit_record_to_community_mock_response

    response = InvenioAPI.submit_record_to_community(sample_host_name,sample_token,sample_community_uuid,sample_record_id)

    assert response.status_code == 200, f"Update Draft Record API failed to raise 200"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_put.assert_called_once()

    #now verify call parameters
    call_args = mock_put.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/records/{sample_record_id}/draft/review"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {sample_token}"
        }

    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

    #check certain key parameters of the body
    called_body: dict[str,Any] = call_args[1]["json"]
    assert called_body.get("type", None) == "community-submission", f"Expected type to be 'community-submission', got {called_body.get('type', None)}"
    assert called_body["receiver"].get("community", None) == sample_community_uuid, f"Expected community UUID to be '{sample_community_uuid}', got {called_body['receiver'].get('community', None)}"

###############################################################################
# Submit Record For Review API
###############################################################################

@pytest.fixture
def create_submit_record_for_review_mock_response(mocker) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 202
    
    #NO PAYLOAD
    mock_response.json.return_value = {}

    mock_post = mocker.patch("requests.post",return_value = mock_response)

    return mock_post,mock_response

@pytest.mark.unit
def test_submit_record_for_review(mocker,create_submit_record_for_review_mock_response,
                    sample_host_name,sample_token,sample_record_id):
    mock_post,mock_response = create_submit_record_for_review_mock_response

    payload = {"content": "THIS IS A COMMENT","format": "html"}
    response = InvenioAPI.submit_record_for_review(sample_host_name,sample_token,sample_record_id,payload)

    assert response.status_code == 202, f"Update Draft Record API failed to raise 202"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_post.assert_called_once()

    #now verify call parameters
    call_args = mock_post.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/records/{sample_record_id}/draft/actions/submit-review"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers =  {"Authorization": f"Bearer {sample_token}"}
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

    #check certain key parameters of the body
    called_body: dict[str,Any] = call_args[1]["json"]
    assert called_body.get("payload", None) == payload, f"Expected payload to be {payload}, got {called_body.get('payload', None)}"

###############################################################################
# Get Draft Record API
###############################################################################

@pytest.fixture
def create_get_draft_record_mock_response(mocker,sample_host_name,sample_record_id,
                                        sample_metadata) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 200
    
    mock_response.json.return_value = {
        "id": sample_record_id,
        "links": {
            "latest": f"{sample_host_name}/api/records/{sample_record_id}/versions/latest",
            "files": f"{sample_host_name}/api/records/{sample_record_id}/draft/files"
        },
        "metadata": sample_metadata,
        "is-published": False
    }

    mock_get = mocker.patch("requests.get",return_value = mock_response)

    return mock_get,mock_response

@pytest.mark.unit
def test_get_draft_record(mocker,create_get_draft_record_mock_response,
                    sample_host_name,sample_token,sample_record_id):
    mock_post,mock_response = create_get_draft_record_mock_response

    response = InvenioAPI.get_draft_record(sample_host_name,sample_token,sample_record_id)

    assert response.status_code == 200, f"Update Draft Record API failed to raise 200"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_post.assert_called_once()

    #now verify call parameters
    call_args = mock_post.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/records/{sample_record_id}/draft"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers =  {"Authorization": f"Bearer {sample_token}"}
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

###############################################################################
# Delete Review Request API
###############################################################################

@pytest.fixture
def create_delete_review_request_mock_response(mocker) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 204
    
    #no content
    mock_response.json.return_value = {}
    mock_delete = mocker.patch("requests.delete",return_value = mock_response)

    return mock_delete,mock_response

@pytest.mark.unit
def test_delete_review_request(mocker,create_delete_review_request_mock_response,
                    sample_host_name,sample_token,sample_record_id):
    mock_delete,mock_response = create_delete_review_request_mock_response

    response = InvenioAPI.delete_review_request(sample_host_name,sample_token,sample_record_id)

    assert response.status_code == 204, f"Update Draft Record API failed to raise 204"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_delete.assert_called_once()

    #now verify call parameters
    call_args = mock_delete.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/records/{sample_record_id}/draft/review"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers =  {"Authorization": f"Bearer {sample_token}"}
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

###############################################################################
# Cancel Review Request API
###############################################################################

@pytest.fixture
def create_cancel_review_request_mock_response(mocker,sample_host_name,sample_request_id) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 200
    
    request_id = sample_request_id
    mock_response.json.return_value = {
        "id": request_id,
        "links": {
            "self": f"{sample_host_name}/api/requests/{request_id}",
            "actions": {
                "accept": f"{sample_host_name}/api/requests/{request_id}/actions/accept",
                "cancel": f"{sample_host_name}/api/requests/{request_id}/actions/cancel",
                "decline": f"{sample_host_name}/api/requests/{request_id}/actions/decline",
                "submit": f"{sample_host_name}/api/requests/{request_id}/actions/submit",
            }
        }
    }
    mock_post = mocker.patch("requests.post",return_value = mock_response)

    return mock_post,mock_response

@pytest.mark.unit
def test_cancel_review_request(mocker,create_cancel_review_request_mock_response,
                    sample_host_name,sample_token,sample_request_id):
    mock_post,mock_response = create_cancel_review_request_mock_response

    payload = {"content": "This is a comment: DELETE", "format": "html"}
    response = InvenioAPI.cancel_review_request(sample_host_name,sample_token,sample_request_id,payload)

    assert response.status_code == 200, f"Update Draft Record API failed to raise 200"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_post.assert_called_once()

    #now verify call parameters
    call_args = mock_post.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/requests/{sample_request_id}/actions/cancel"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers =  {"Authorization": f"Bearer {sample_token}"}
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"

    #check certain key parameters of the body
    called_body: dict[str,Any] = call_args[1]["json"]
    assert called_body.get("payload", None) == payload, f"Expected payload to be {payload}, got {called_body.get('payload', None)}"

###############################################################################
# Get (Published) Record Request API
###############################################################################

@pytest.fixture
def create_get_record_mock_response(mocker,sample_host_name,sample_record_id,
                                        sample_metadata) -> Any:
    mock_response: requests.Response = mocker.MagicMock(spec=requests.Response)
    mock_response.status_code = 200
    
    mock_response.json.return_value = {
        "id": sample_record_id,
        "links": {
            "latest": f"{sample_host_name}/api/records/{sample_record_id}/versions/latest",
            "files": f"{sample_host_name}/api/records/{sample_record_id}/draft/files"
        },
        "metadata": sample_metadata,
        "is_published": True
    }

    mock_get = mocker.patch("requests.get",return_value = mock_response)

    return mock_get,mock_response

@pytest.mark.unit
def test_get_draft_record(mocker,create_get_draft_record_mock_response,
                    sample_host_name,sample_token,sample_record_id):
    mock_post,mock_response = create_get_draft_record_mock_response

    response = InvenioAPI.get_record(sample_host_name,sample_token,sample_record_id)

    assert response.status_code == 200, f"Update Draft Record API failed to raise 200"
    assert response.json() == mock_response.json.return_value, f"Failed to retrieve payload"
    mock_post.assert_called_once()

    #now verify call parameters
    call_args = mock_post.call_args

    #Check correct url is constructed
    expected_url = sample_host_name + f"/api/records/{sample_record_id}"
    called_url = call_args[0][0]
    assert called_url == expected_url, f"Expected URL: {expected_url}, Actual URL: {called_url}"

    #Check correct header is passed
    called_headers = call_args[1]['headers']
    expected_headers =  {"Authorization": f"Bearer {sample_token}"}
    assert called_headers == expected_headers, f"Expected headers: {expected_headers}, Actual headers: {called_headers}"
