import pytest
import h5py
import numpy as np
from NekUpload.utils.extractor import HDF5Extractor

# Note that tmp_path is from pytest, allows automatic cleanup of generated files

###############################################################################
# TEST HDF5Extractor
###############################################################################

@pytest.mark.unit
def test_hdf5_attr_extractor(tmp_path):
    filename = tmp_path / "testfile.h5" #pytest will handle cleanup

    # Create a new HDF5 file
    with h5py.File(filename, 'w') as f:
        grp = f.create_group('test_group')
        grp.attrs['test_attribute'] = 'test_value'

    #read file and test attribute extraction
    with h5py.File(filename,'r') as f:
        val = HDF5Extractor.extract_attribute(f,"test_group","test_attribute")
        assert val == "test_value", f"File should have attribute 'test_attribute' with value 'test_value', got {val}"

@pytest.mark.unit
def test_hdf5_attr_extractor_bad_group(tmp_path):
    filename = tmp_path / "testfile.h5"

    # Create a new HDF5 file
    with h5py.File(filename, 'w') as f:
        grp = f.create_group('test_group')
        grp.attrs['test_attribute'] = 'test_value'

    #read file and test attribute extraction
    with h5py.File(filename,'r') as f:
        val = HDF5Extractor.extract_attribute(f,"invalid_group","test_attribute")
        assert not val, f"Group does not exist, so should get None, got {val}"

@pytest.mark.unit
def test_hdf5_attr_extractor_bad_attribute(tmp_path):
    filename = tmp_path / "testfile.h5"

    # Create a new HDF5 file
    with h5py.File(filename, 'w') as f:
        grp = f.create_group('test_group')
        grp.attrs['test_attribute'] = 'test_value'

    with h5py.File(filename,'r') as f:
        val = HDF5Extractor.extract_attribute(f,"test_group","invalid_attribute")
        assert not val, f"Attribute does not exist, so should get None, got {val}"

@pytest.mark.unit
def test_hdf5_attr_geometry_streaming(tmp_path):
    filename = tmp_path / "testfile.h5"

    # Create a new HDF5 file
    with h5py.File(filename, 'w') as f:
        grp = f.create_group('test_group')

        # Create a dataset with 1000 coordinates with random numbers between 0 and 100
        data = np.random.randint(0, 100, size=(1000, 3))
        
        #make sure min is 0 and max is 100
        data[400][0] = 0
        data[304][1] = 0
        data[999][2] = 0

        data[123][0] = 100
        data[2][1] = 100
        data[555][2] = 100

        grp.create_dataset('coordinates', data=data)

    with h5py.File(filename,'r') as f:
        min_coord,max_coord = HDF5Extractor.extract_min_max_coords(f,"test_group/coordinates")
        assert np.array_equal(min_coord, np.array([0, 0, 0])), f"Expected min coordinates to be [0, 0, 0], but got {min_coord}"
        assert np.array_equal(max_coord, np.array([100, 100, 100])), f"Expected max coordinates to be [100, 100, 100], but got {max_coord}"

###############################################################################
# TEST NekAutoExtractor
###############################################################################
