import pytest
import h5py
import os
import numpy as np
from NekUpload.utils.hdf5_reader import HDF5Reader,HDF5ReaderException

###############################################################################
# Set up fixtures here
# Using tmp_path allows for auto file cleanup
###############################################################################

@pytest.fixture
def setup_hdf5_file(tmp_path):
    file_path = tmp_path / 'test_file.h5'

    with h5py.File(file_path, 'a') as f:
        group1 = f.create_group('ROOT/GROUP1')
        group2 = f.create_group('ROOT/GROUP2')

        #group 1 should have attributes
        data1 = np.random.random((100, 100))
        dataset1 = group1.create_dataset("dataset_with_attrs", data=data1)
        dataset1.attrs["description"] = "This dataset has attributes"
        dataset1.attrs["scale"] = 10

        #group 2 has no attributes
        data2 = np.array([10,20,30,40,50])
        group2.create_dataset("dataset_no_attrs", data=data2)

    return file_path,data1,data2

###############################################################################
# Test HDF5Reader here
###############################################################################

@pytest.mark.unit
def test_valid_get_keys(setup_hdf5_file):
    file_path,*_ = setup_hdf5_file
    with HDF5Reader(file_path) as reader:
        keys = reader.get_keys()
        expected_keys = {
            "ROOT": "GROUP",
            "ROOT/GROUP1": "GROUP",
            "ROOT/GROUP1/dataset_with_attrs": "DATASET",
            "ROOT/GROUP1/dataset_with_attrs/@description": "ATTRIBUTE",
            "ROOT/GROUP1/dataset_with_attrs/@scale": "ATTRIBUTE",
            "ROOT/GROUP2": "GROUP",
            "ROOT/GROUP2/dataset_no_attrs": "DATASET"
        }
        assert keys == expected_keys

@pytest.mark.unit
def test_invalid_get_keys(setup_hdf5_file):
    file_path,*_ = setup_hdf5_file

    #this is not correct way to use this class, should use as context manager
    reader = HDF5Reader(file_path)
    with pytest.raises(HDF5ReaderException):
        reader.get_keys()

@pytest.mark.unit
def test_valid_get_dataset(setup_hdf5_file):
    file_path,data1,data2 = setup_hdf5_file
    with HDF5Reader(file_path) as reader:
        dataset1 = reader.get_dataset("ROOT/GROUP1/dataset_with_attrs")
        dataset2 = reader.get_dataset("ROOT/GROUP2/dataset_no_attrs")
        
        np.testing.assert_array_equal(dataset1[...], data1)
        np.testing.assert_array_equal(dataset2[...], data2)

@pytest.mark.unit
def test_invalid_get_dataset(setup_hdf5_file):
    file_path,_,_ = setup_hdf5_file

    with HDF5Reader(file_path) as f:
        with pytest.raises(HDF5ReaderException):
            f.get_dataset("THIS/DATASET/NOT/EXISTS")

@pytest.mark.unit
def test_valid_get_attributes(setup_hdf5_file):
    file_path,*_ = setup_hdf5_file

    with HDF5Reader(file_path) as reader:
        attrs1 = reader.get_attributes("ROOT/GROUP1/dataset_with_attrs")
        attrs2 = reader.get_attributes("ROOT/GROUP2/dataset_no_attrs")
        
        expected_attrs1 = {
            "description": "This dataset has attributes",
            "scale": 10
        }
        expected_attrs2 = {}

        assert attrs1 == expected_attrs1
        assert attrs2 == expected_attrs2

@pytest.mark.unit
def test_invalid_get_attributes(setup_hdf5_file):
    file_path,*_ = setup_hdf5_file

    with HDF5Reader(file_path) as reader:
        with pytest.raises(HDF5ReaderException):
            reader.get_attributes("THIS/DOES/NOT/EXIST")

@pytest.mark.unit
def test_dump_to_plain_file(setup_hdf5_file,tmp_path):
    file_path,*_ = setup_hdf5_file
    
    output_file = tmp_path / "output.txt"
    with HDF5Reader(file_path) as reader:
        reader.dump_to_plain_file(output_file)
        
    assert os.path.exists(output_file), "Output file was not created."
    
    with open(output_file, "r") as f:
        content = f.read()
        assert "Group: /ROOT/GROUP1" in content, "Group /ROOT/GROUP1 not found in output file."
        assert "Dataset: dataset_with_attrs" in content, "Dataset dataset_with_attrs not found in output file."
        assert "Attributes:" in content, "Attributes section not found in output file."
        assert "description: This dataset has attributes" in content, "Attribute 'description' not found in output file."
        assert "scale: 10" in content, "Attribute 'scale' not found in output file."
        assert "Group: /ROOT/GROUP2" in content, "Group /ROOT/GROUP2 not found in output file."
        assert "Dataset: dataset_no_attrs" in content, "Dataset dataset_no_attrs not found in output file."

@pytest.mark.unit
def test_summary(setup_hdf5_file):
    file_path,*_ = setup_hdf5_file

    with HDF5Reader(file_path) as reader:
        summary = reader.summary()
        
        assert summary["File Name"] == file_path, f"Expected file name {file_path}, but got {summary['File Name']}"
        assert summary["Groups"] == 3, f"Expected 3 groups, but got {summary['Groups']}"
        assert summary["Datasets"] == 2, f"Expected 2 datasets, but got {summary['Datasets']}"
        assert summary["Attributes"] == 2, f"Expected 2 attributes, but got {summary['Attributes']}"

@pytest.mark.unit
def test_valid_get_dtype(setup_hdf5_file):
    file_path,data1,data2 = setup_hdf5_file

    expected_dtype1 = data1.dtype
    expected_dtype2 = data2.dtype

    with HDF5Reader(file_path) as reader:
        dtype1 = reader.get_dtype("ROOT/GROUP1/dataset_with_attrs")
        dtype2 = reader.get_dtype("ROOT/GROUP2/dataset_no_attrs")
        
        assert dtype1 == str(expected_dtype1)
        assert dtype2 == str(expected_dtype2)

@pytest.mark.unit
def test_invalid_get_dtype(setup_hdf5_file):
    file_path,*_ = setup_hdf5_file

    with HDF5Reader(file_path) as reader:
        with pytest.raises(HDF5ReaderException):
            reader.get_dtype("THIS/DATASET/DOES/NOT/EXIST")

@pytest.mark.unit
def test_valid_get_shape(setup_hdf5_file):
    file_path,*_ = setup_hdf5_file

    expected_shapes_group_1 = (100,100)
    expected_shapes_group_2 = (5,)

    with HDF5Reader(file_path) as reader:
        shape1 = reader.get_shape("ROOT/GROUP1/dataset_with_attrs")
        shape2 = reader.get_shape("ROOT/GROUP2/dataset_no_attrs")
        
        assert shape1 == expected_shapes_group_1, f"Expected shape {expected_shapes_group_1}, but got {shape1}"
        assert shape2 == expected_shapes_group_2, f"Expected shape {expected_shapes_group_2}, but got {shape2}"

@pytest.mark.unit
def test_invalid_get_shape(setup_hdf5_file):
    file_path,*_ = setup_hdf5_file

    with HDF5Reader(file_path) as reader:
        with pytest.raises(HDF5ReaderException):
            reader.get_shape("THIS/DATASET/NO/EXISTO")