import h5py
import pytest
from NekUpload.utils import parsing

@pytest.mark.unit
def test_split_equals_expr():
    expr = "abc def = 1234    "
    lhs = "abc def"
    rhs = "1234"

    expr_lhs,expr_rhs = parsing.get_both_sides_of_equals(expr)

    assert expr_lhs == lhs, f"Expected lhs to be '{lhs}', but got '{expr_lhs}'"
    assert expr_rhs == rhs, f"Expected rhs to be '{rhs}', but got '{expr_rhs}'"

@pytest.mark.unit
@pytest.mark.parametrize(
    "description, data, expected",
    [
        (
            "no-evaluation",
            {"PARAM1": "10", "PARAM2": "20"},
            {"PARAM1": 10, "PARAM2": 20},
        ),
        (
            "sub-only",
            {"PARAM1": "10", "PARAM2": "PARAM1"},
            {"PARAM1": 10, "PARAM2": 10},
        ),
        (
            "sub-with-literals",
            {"PARAM1": "10", "PARAM2": "PARAM1 + 5"},
            {"PARAM1": 10, "PARAM2": 15},
        ),
        (
            "sub-with-vars",
            {"PARAM1": "10", "PARAM2": "2", "PARAM3": "PARAM1 * PARAM2"},
            {"PARAM1": 10, "PARAM2": 2, "PARAM3": 20},
        ),
    ]
)
def test_resolve_parameters(description: str,data: dict[str,str],expected:dict[str,float|int]):
    """Test evaluation of paramater map. Add the following to the parameter map
    to extend tests:

    Args:
        description (str): Description of the test
        data (dict[str,str]): Mapping of parameters to values in string, allows cross-references
        expected (dict[str, float | int]): Mapping of parameters to the numeric value
    """

    result = parsing.evaluate_parameters(data)
    assert result == expected, f"Test {description} expected result to be '{expected}', but got '{result}'"

@pytest.mark.unit
@pytest.mark.parametrize(
    "description, files, extension, final_files",
    [
        (
            "normal",
            ["a.py", "b.py", "c.xml", "d.xml", "e.py"],
            ".py",
            ["a.py", "b.py", "e.py"],
        ),
        (
            "case_sensitive",
            ["a.py", "b.py", "c.xml", "d.XML", "e.py"],
            ".xml",
            ["c.xml", "d.XML"],
        ),
        (
            "missing_dot_ext",
            ["a.py", "b.py", "c.xml", "d.XML", "e.py"],
            "xml",
            ["c.xml", "d.XML"],
        ),
        (
            "empty",
            ["a.py", "b.py", "c.xml", "d.XML", "e.py"],
            ".rst",
            [],
        ),
    ],
)
def test_get_all_file_extensions(description:str,files:list[str],extension:str,final_files:list[str]):
    """Test get all file extensions method. To add more tests, specify in the parametrization

    Args:
        description (str): Descriptor for what you are testing
        files (list[str]): List of files
        extension (str): Extension to acquire
        final_files (list[str]): Filtered list of files containing the extension
    """
    get_files = parsing.get_all_files_with_extension(files,extension)

    assert get_files == final_files, f"Test {description} expected {final_files}, but got {get_files}" 

@pytest.mark.unit
def test_get_hdf5_groups_one_level_depth(valid_geometry_HDF5_files):
    nekg_files = valid_geometry_HDF5_files
    expected_root_groups = ["","NEKTAR"] #HDF5 files have a parent directory where everythin is derived from

    for hdf5_file in nekg_files:
        with h5py.File(hdf5_file) as f:
            groups: list[str] = parsing.get_hdf5_groups_with_depth_limit(f,1)
            assert groups == expected_root_groups

@pytest.mark.unit
def test_get_hdf5_groups_multi_level_depth(valid_geometry_HDF5_files):
    nekg_files = valid_geometry_HDF5_files
    #HDF5 files have a parent directory where everythin is derived from
    expected_root_groups = ["","NEKTAR","NEKTAR/GEOMETRY","NEKTAR/GEOMETRY/MAPS","NEKTAR/GEOMETRY/MESH"] 

    for hdf5_file in nekg_files:
        with h5py.File(hdf5_file) as f:
            groups: list[str] = parsing.get_hdf5_groups_with_depth_limit(f,5)
            assert groups == expected_root_groups

@pytest.mark.unit
def test_get_hdf5_groups_multi_level_depth_new_start_point(valid_geometry_HDF5_files):
    nekg_files = valid_geometry_HDF5_files
    expected_groups = ["NEKTAR/GEOMETRY","NEKTAR/GEOMETRY/MAPS","NEKTAR/GEOMETRY/MESH"] 

    for hdf5_file in nekg_files:
        with h5py.File(hdf5_file) as f:
            groups: list[str] = parsing.get_hdf5_groups_with_depth_limit(f,1,"NEKTAR/GEOMETRY")
            assert groups == expected_groups

@pytest.mark.unit
def test_get_hdf5_groups_test_limit(valid_geometry_HDF5_files):
    nekg_files = valid_geometry_HDF5_files

    #any more, and difficult to test as each geometry file can have different number of groups
    #but all at minimum have these groups
    all_groups = ["","NEKTAR/GEOMETRY","NEKTAR/GEOMETRY/MAPS","NEKTAR/GEOMETRY/MESH"]

    max_groups = range(0,5)
    expected_nums = [min(max_num,len(all_groups)) for max_num in max_groups]

    for hdf5_file in nekg_files:
        with h5py.File(hdf5_file) as f:
            for max_num,expected_num in zip(max_groups,expected_nums):
                groups: list[str] = parsing.get_hdf5_groups_with_depth_limit(f,3,"NEKTAR",max_num)
                assert expected_num == len(groups)

@pytest.mark.unit
def test_get_hdf5_datasets_one_level_depth(valid_geometry_HDF5_files):
    nekg_files = valid_geometry_HDF5_files
    possible_datasets = [] #there are no datasets in first level of NEKTAR 

    for hdf5_file in nekg_files:
        with h5py.File(hdf5_file) as f:
            datasets: list[str] = parsing.get_hdf5_datasets_with_depth_limit(f,1)
            assert possible_datasets == datasets

@pytest.mark.unit
def test_get_hdf5_datasets_multi_level_depth(valid_geometry_HDF5_files,geometry_possible_datasets):
    nekg_files = valid_geometry_HDF5_files

    possible_datasets: list[str] = geometry_possible_datasets
    
    for hdf5_file in nekg_files:
        with h5py.File(hdf5_file) as f:
            datasets: list[str] = parsing.get_hdf5_datasets_with_depth_limit(f,3,"NEKTAR")
            
            #geometry file doesn't have all 3d elements if not used
            #so just make sure all files are allowable
            for dataset in datasets:
                assert dataset in possible_datasets

@pytest.mark.unit
def test_get_hdf5_datasets_test_limit(valid_geometry_HDF5_files):
    nekg_files = valid_geometry_HDF5_files

    #any more, and difficult to test as each geometry file can have different number of datasets
    #but all at minimum have these datasets    
    max_datasets = [0,1,2,3,4,5,6,7,8]  

    for hdf5_file in nekg_files:
        with h5py.File(hdf5_file) as f:
            for max_num in max_datasets:
                datasets: list[str] = parsing.get_hdf5_datasets_with_depth_limit(f,3,"NEKTAR",max_num)
                assert max_num == len(datasets)