from NekUpload.utils.xml_reader import XMLReader
import os
import pytest
import lxml.etree as ET

###############################################################################
# Set up some useful files for testing, in form of fixtures
# Using tmp_path from pytest allows for automatic file deletion 
###############################################################################

@pytest.fixture
def xml_file_1_structure():
    return [
        "NEKTAR",
        "NEKTAR/REPLACE",
        "NEKTAR/REPLACE/NESTED",
        "NEKTAR/NOREPLACE",
        "NEKTAR/NOREPLACE/TAG1"
    ]

@pytest.fixture
def create_xml_file_1(tmp_path):
    xml_content = """<?xml version="1.0" encoding="utf-8" ?>
<NEKTAR>
    <REPLACE>
        <NESTED>REPLACED</NESTED>
    </REPLACE>

    <NOREPLACE>
        <TAG1 VALUE="1"/>
    </NOREPLACE>
</NEKTAR>
"""
    filename = tmp_path / 'file1.xml'
    filename.write_text(xml_content,encoding="utf-8")

    return filename

@pytest.fixture
def create_xml_file_2(tmp_path):
    xml_content = """<?xml version="1.0" encoding="utf-8" ?>
<NEKTAR>
    <REPLACE>
        <NESTED>NEW REPLACED</NESTED>
    </REPLACE>

    <EXTRA>
        <TAG>
            <VAL value="1"/>
        </TAG>
    </EXTRA>
</NEKTAR>
"""
    filename = tmp_path / 'file2.xml'
    filename.write_text(xml_content,encoding="utf-8")

    return filename

@pytest.fixture
def create_xml_file_1_2_expected_merge(tmp_path):
    xml_content = """<NEKTAR>
    <REPLACE>
        <NESTED>NEW REPLACED</NESTED>
    </REPLACE>
    <NOREPLACE>
        <TAG1 VALUE="1"/>
    </NOREPLACE>
    <EXTRA>
        <TAG>
            <VAL value="1"/>
        </TAG>
    </EXTRA>
</NEKTAR>
"""
    filename = tmp_path / 'expected_merge.xml'
    filename.write_text(xml_content,encoding="utf-8")

    return filename

@pytest.fixture
def change_test_dir():
    original_dir = os.getcwd()
    test_dir = os.path.dirname(__file__)
    os.chdir(test_dir)
    yield
    os.chdir(original_dir)

###############################################################################
# Helper functions here
###############################################################################
def xml_files_are_equal(file1: str, file2: str) -> bool:
    """Compares two XML files for content equality, ignoring formatting differences."""
    tree1 = ET.parse(file1)
    tree2 = ET.parse(file2)

    # Convert both XMLs into a canonicalized string format
    xml1 = ET.tostring(tree1, method="c14n")
    xml2 = ET.tostring(tree2, method="c14n")

    return xml1 == xml2

###############################################################################
# Actual tests here
###############################################################################

@pytest.mark.unit
def test_merge(tmp_path,create_xml_file_1,create_xml_file_2,create_xml_file_1_2_expected_merge):
    file1 = create_xml_file_1
    file2 = create_xml_file_2
    expected_merge_file = create_xml_file_1_2_expected_merge

    output_file = tmp_path / "output.xml"

    with XMLReader(file1) as f:
        f.merge_first_level_elements_with(file2,output_file)

    #may want to check formatting at some point in future, but currently looks good
    assert xml_files_are_equal(expected_merge_file,output_file)

@pytest.mark.unit
def test_get_structure(create_xml_file_1,xml_file_1_structure):
    file1 = create_xml_file_1
    expected_xml_structure = xml_file_1_structure

    with XMLReader(file1) as f:
        structure: list[str] = f.get_structure()

    assert structure == expected_xml_structure
