import pytest
import h5py
import numpy as np

from NekUpload.validate.exceptions import SessionFileException
from NekUpload.validate.files import NekGeometryFile,NekSessionFile
from NekUpload.NekData.composite import CompositeDefinition
from NekUpload.NekData.expansions import ExpansionDefinition
from NekUpload.NekData.filters import *
from NekUpload.testutils import GEOMETRY_COMPOSITE_DATASETS,SESSION_EXPANSION_DATASETS
from NekUpload.testutils import NekTestDataset,NekTestGeometryComposite,NekTestSessionExpansion

###############################################################################
# Test GeometryFile here
###############################################################################
@pytest.mark.unit
@pytest.mark.parametrize("test_case",GEOMETRY_COMPOSITE_DATASETS,ids=lambda dataset: str(dataset[0]))
def test_get_composite_data(test_case: tuple[NekTestDataset,NekTestGeometryComposite]):
    dataset: NekTestDataset = test_case[0]
    expected_composite_info: NekTestGeometryComposite = test_case[1]
    geometry: str = dataset.geometry

    expected_data_raw: dict[int,str] = expected_composite_info.composite_id_to_definition_map
    expected_data: dict[int,CompositeDefinition] = {idx:CompositeDefinition(data) for idx,data in expected_data_raw.items()}

    with NekGeometryFile(geometry) as f:
        composite_info = f.get_composite_info()

    assert composite_info == expected_data, f"Expected {expected_data}, but got {composite_info}"

@pytest.mark.unit
@pytest.mark.parametrize("test_case",
                        [info for info in GEOMETRY_COMPOSITE_DATASETS if info[1].dim == 2],#only 2d geometries contain edges
                        ids=lambda dataset: str(dataset[0]))
def test_get_composite_edges_only_data(test_case: tuple[NekTestDataset,NekTestGeometryComposite]):
    dataset: NekTestDataset = test_case[0]
    expected_composite_info: NekTestGeometryComposite = test_case[1]
    geometry: str = dataset.geometry

    expected_data_raw: dict[int,str] = expected_composite_info.boundary_composite_id_to_definition_map
    expected_data: dict[int,CompositeDefinition] = {idx:CompositeDefinition(data) for idx,data in expected_data_raw.items()}

    with NekGeometryFile(geometry) as f:
        composite_info = f.get_composite_edge_info()

    assert composite_info == expected_data, f"Expected {expected_data}, but got {composite_info}"

@pytest.mark.unit
@pytest.mark.parametrize("test_case",
                        [info for info in GEOMETRY_COMPOSITE_DATASETS if info[1].dim == 3],#only 3d geometries contain edges
                        ids=lambda dataset: str(dataset[0]))
def test_get_composite_faces_only_data(test_case: tuple[NekTestDataset,NekTestGeometryComposite]):
    dataset: NekTestDataset = test_case[0]
    expected_composite_info: NekTestGeometryComposite = test_case[1]
    geometry: str = dataset.geometry

    expected_data_raw: dict[int,str] = expected_composite_info.boundary_composite_id_to_definition_map
    expected_data: dict[int,CompositeDefinition] = {idx:CompositeDefinition(data) for idx,data in expected_data_raw.items()}

    with NekGeometryFile(geometry) as f:
        composite_info = f.get_composite_face_info()

    assert composite_info == expected_data, f"Expected {expected_data}, but got {composite_info}"

###############################################################################
# Test SessionFile here
###############################################################################

@pytest.fixture
def write_boundary_session_info_file(tmp_path):
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <BOUNDARYREGIONS>
                <B ID="0">C[2]</B>
                <B ID="1">C[3]</B>
                <B ID="2">C[4]</B>
                <B ID="3">C[5]</B>
                <B ID="4">C[6]</B>
                <B ID="5">C[7]</B>
                <B ID="6">C[8]</B>
                <B ID="7">C[9]</B>
                <B ID="8">C[10]</B>
            </BOUNDARYREGIONS>
            
            <BOUNDARYCONDITIONS>
                <REGION REF="0">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="1">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="2">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="3">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="4">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="5">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="6">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="7">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="8">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="9">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="10">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
            </BOUNDARYCONDITIONS>
        </CONDITIONS>
    </NEKTAR>
    '''    
    
    # Write to a file
    filename = tmp_path / "test_conditions.xml"
    with open(filename, 'w') as file:
        file.write(xml_content)

    return str(filename)

@pytest.fixture
def write_boundary_session_info_insufficient_boundaries_file(tmp_path):
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <BOUNDARYREGIONS>
                <B ID="0">C[2]</B>
                <B ID="1">C[3]</B>
                <B ID="2">C[4]</B>
                <B ID="3">C[5]</B>
                <B ID="4">C[6]</B>
                <B ID="5">C[7]</B>
                <B ID="6">C[8]</B>
                <B ID="7">C[9]</B>
                <B ID="8">C[10]</B>
            </BOUNDARYREGIONS>
            
            <BOUNDARYCONDITIONS>
                <REGION REF="0">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="1">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
            </BOUNDARYCONDITIONS>
        </CONDITIONS>
    </NEKTAR>
    '''

    # Write to a file
    filename = tmp_path / "test_conditions_insufficiently_defined.xml"
    with open(filename, 'w') as file:
        file.write(xml_content)

    return str(filename)

@pytest.mark.unit
def test_boundary_session_info(write_boundary_session_info_file):
    expected_data_raw = {
        0: "C[2]",
        1: "C[3]",
        2: "C[4]",
        3: "C[5]",
        4: "C[6]",
        5: "C[7]",
        6: "C[8]",
        7: "C[9]",
        8: "C[10]",
    }
    expected_data = {idx: CompositeDefinition(data) for idx,data in expected_data_raw.items()}

    with NekSessionFile(write_boundary_session_info_file) as f:
        boundary_info = f.get_all_defined_boundary_regions()

    assert boundary_info == expected_data, f"Expected {expected_data}, but got {boundary_info}"

@pytest.mark.unit
def test_boundary_session_insufficient_boundaries_info(write_boundary_session_info_insufficient_boundaries_file):
    with NekSessionFile(write_boundary_session_info_insufficient_boundaries_file) as f:
        with pytest.raises(SessionFileException):
            f.get_all_defined_boundary_regions()

@pytest.fixture
def write_geometry_info_file(tmp_path):
    xml_content = f'''<NEKTAR>
                <GEOMETRY DIM="2" SPACE="2" HDF5FILE="TEST.nekg" />
    </NEKTAR>
    '''

    # Write to a file
    filename = tmp_path / "geometry_info_only.xml"
    with open(filename, 'w') as file:
        file.write(xml_content)

    return str(filename)

@pytest.mark.unit
def test_geometry_info(write_geometry_info_file):
    expected_elmt_dim = 2
    expected_space_dim = 2
    expected_hdf5file = "TEST.nekg"
    
    with NekSessionFile(write_geometry_info_file) as f:
        elmt_dim,space_dim,hdf5file = f.get_geometry_info()

    assert elmt_dim == expected_elmt_dim
    assert space_dim == expected_space_dim, f"Expected space dimension {expected_space_dim}, but got {space_dim}"
    assert hdf5file == expected_hdf5file, f"Expected HDF5 file {expected_hdf5file}, but got {hdf5file}"

@pytest.mark.unit
def test_get_boundary_conditions_none(tmp_path):
    #set up test file with no files referenced
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <BOUNDARYCONDITIONS>
                <REGION REF="0">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="1">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
            </BOUNDARYCONDITIONS>
        </CONDITIONS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "geometry_info_only.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    with NekSessionFile(filename) as f:
        file_list = f.get_all_boundary_condition_files()
        expected = []
        assert expected == file_list, f"Expected {expected}, but got {file_list}"
    
@pytest.mark.unit
def test_get_boundary_conditions_one(tmp_path):
    #set up test file with no files referenced
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <BOUNDARYCONDITIONS>
                <REGION REF="0">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="1">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" FILE="test.bc" />
                </REGION>
            </BOUNDARYCONDITIONS>
        </CONDITIONS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "geometry_info_only.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    with NekSessionFile(filename) as f:
        file_list = f.get_all_boundary_condition_files()
        expected = ["test.bc"]
        assert expected == file_list, f"Expected {expected}, but got {file_list}"
    
@pytest.mark.unit
def test_get_boundary_conditions_two(tmp_path):
    #set up test file with no files referenced
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <BOUNDARYCONDITIONS>
                <REGION REF="0">
                    <D VAR="u" FILE="test1.bc" />
                </REGION>
                <REGION REF="1">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" FILE="test2.bc" />
                </REGION>
            </BOUNDARYCONDITIONS>
        </CONDITIONS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "geometry_info_only.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    with NekSessionFile(filename) as f:
        file_list = f.get_all_boundary_condition_files()
        expected = ["test1.bc","test2.bc"]
        #make order redundant
        assert set(expected) == set(file_list), f"Expected {expected}, but got {file_list}"

###############################################################################
# Test parsing of PARAMETERS in CONDITIONS
###############################################################################

@pytest.mark.unit
def test_get_parameters_normal(tmp_path):
    #set up test file with no files referenced
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <PARAMETERS>
                <P>TimeStep = 2.5e-4</P>
                <P>NumSteps = 20</P>
                <P>IO_CheckSteps = 1000</P>
                <P>IO_InfoSteps = 1</P>
                <P>IO_FiltersInfoSteps = 100</P>
                <P>advx = 1</P>
                <P>advy = 0</P>
                <P>epsilon = 0.000001</P>
                <P>GJPJumpScale = 1</P>
            </PARAMETERS>
        </CONDITIONS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "get_params_easy.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    #now test file
    with NekSessionFile(filename) as f:
        params: dict[str,float | int] = f.get_parameters()
    
    expected_params = {
        "TimeStep": 2.5e-4,
        "NumSteps": 20,
        "IO_CheckSteps": 1000,
        "IO_InfoSteps": 1,
        "IO_FiltersInfoSteps": 100,
        "advx": 1,
        "advy": 0,
        "epsilon": 0.000001,
        "GJPJumpScale": 1,
    }

    assert len(params) == len(expected_params)

    for key, expected_value in expected_params.items():
        if isinstance(expected_value, float):
            assert params[key] == pytest.approx(expected_value, rel=1e-6)
        else:
            assert params[key] == expected_value

@pytest.mark.unit
def test_get_parameters_hard(tmp_path):
    #set up test file with no files referenced
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <PARAMETERS>
                <P> TimeStep              = 0.05                          </P>
                <P> NumSteps              = 30                           </P>
                <P> IO_CheckSteps         = 0                             </P>
                <P> IO_InfoSteps          = 1                             </P>
                <P> Gamma                 = 1.4                           </P>
                <P> GasConstant           = 1/1.4                         </P>
                <P> cInf                  = 1                             </P>
                <P> rhoInf                = 1                             </P>
                <P> Mach                  = 0.5                           </P>
                <P> Re                    = 100                           </P>
                <P> Pr                    = 0.72                          </P>
                <P> L                     = 1                             </P>
                <P> NonlinIterTolRelativeL2   = 1.0E-10                   </P>
                <P> NewtonRelativeIteTol      = 1.0E-10                   </P>
                <P> LinSysRelativeTolInNonlin = 1.0E-10                   </P>
                <P> NekNonlinSysMaxIterations = 10                        </P>
                <P> NekLinSysMaxIterations    = 50                        </P>
                <P> PreconMatFreezNumb        = 100                       </P>
                <P> PreconItsStep             = 7                         </P>
                <P> TInf                  = cInf ^ 2 / Gamma / GasConstant</P>
                <P> T0                    = 0.9*TInf                      </P>
                <P> pInf                  = rhoInf * GasConstant * TInf   </P>
                <P> uInf                  = Mach*cInf                     </P>
                <P> vInf                  = 0.0                           </P>
                <P> mu                    = Mach / Re                     </P>
                <P> IPSymmFluxCoeff       = 0.0                           </P>
            </PARAMETERS>
        </CONDITIONS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "get_params_hard.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    #now test file
    with NekSessionFile(filename) as f:
        params: dict[str,float | int] = f.get_parameters()
    
    expected_params = {
        "TimeStep": 0.05,
        "NumSteps": 30,
        "IO_CheckSteps": 0,
        "IO_InfoSteps": 1,
        "Gamma": 1.4,
        "GasConstant": 1 / 1.4,
        "cInf": 1,
        "rhoInf": 1,
        "Mach": 0.5,
        "Re": 100,
        "Pr": 0.72,
        "L": 1,
        "NonlinIterTolRelativeL2": 1.0E-10,
        "NewtonRelativeIteTol": 1.0E-10,
        "LinSysRelativeTolInNonlin": 1.0E-10,
        "NekNonlinSysMaxIterations": 10,
        "NekLinSysMaxIterations": 50,
        "PreconMatFreezNumb": 100,
        "PreconItsStep": 7,
        "TInf": (1 ** 2) / 1.4 / (1 / 1.4),
        "T0": 0.9 * ((1 ** 2) / 1.4 / (1 / 1.4)),
        "pInf": 1 * (1 / 1.4) * ((1 ** 2) / 1.4 / (1 / 1.4)),
        "uInf": 0.5 * 1,
        "vInf": 0.0,
        "mu": 0.5 / 100,
        "IPSymmFluxCoeff": 0.0,
    }

    assert len(params) == len(expected_params)

    for key, expected_value in expected_params.items():
        if isinstance(expected_value, float):
            assert params[key] == pytest.approx(expected_value, rel=1e-6)
        else:
            assert params[key] == expected_value

###############################################################################
# Test extraction of filters
###############################################################################

@pytest.mark.unit
def test_no_filter_no_exception(tmp_path):
    #set up test file with no files referenced
    xml_content = '''<NEKTAR>
        <!-- No filters -->
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "checkpoint_filter.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    #no exceptions should be thrown, should just return empty list
    with NekSessionFile(filename) as f:
        filters: list[CheckpointFilter] = f.get_filters()

    assert filters == [], f"Expected no filters, but got {filters}"

@pytest.mark.unit
def test_checkpoint_filter_one(tmp_path):
    #set up test file with no files referenced
    xml_content = '''<NEKTAR>
        <FILTERS>
            <FILTER TYPE="Checkpoint"> 
                <PARAM NAME="OutputFile">IntermediateFields</PARAM> 
                <PARAM NAME="OutputFrequency">100</PARAM> 
                <PARAM NAME="OutputStartTime"> 50.0 </PARAM> 
            </FILTER>
        </FILTERS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "checkpoint_filter.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    #test here
    with NekSessionFile(filename) as f:
        checkpoint_filters: list[CheckpointFilter] = f.get_filters("Checkpoint")

    assert len(checkpoint_filters) == 1
    assert checkpoint_filters[0].output_file == "IntermediateFields"
    assert checkpoint_filters[0].output_frequency == 100
    assert checkpoint_filters[0].output_start_time == 50.0

@pytest.mark.unit
def test_checkpoint_filter_one_missing_fields(tmp_path):
    #set up test file with no files referenced
    xml_content = '''<NEKTAR>
        <FILTERS>
            <FILTER TYPE="Checkpoint"> 
                <PARAM NAME="OutputFrequency">100</PARAM> 
            </FILTER>
        </FILTERS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "checkpoint_filter.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    #test here
    with NekSessionFile(filename) as f:
        checkpoint_filters: list[CheckpointFilter] = f.get_filters("Checkpoint")

    assert len(checkpoint_filters) == 1
    assert checkpoint_filters[0].output_file == "checkpoint_filter.xml"
    assert checkpoint_filters[0].output_frequency == 100
    assert checkpoint_filters[0].output_start_time == 0.0

@pytest.mark.unit
def test_checkpoint_filter_one_missing_fields(tmp_path):
    #set up test file with no files referenced
    xml_content = '''<NEKTAR>
        <FILTERS>
            <FILTER TYPE="Checkpoint"> 
                <PARAM NAME="OutputFrequency">100</PARAM> 
            </FILTER>
            <FILTER TYPE="HistoryPoints"> <!-- This should be ignored silently -->
            </FILTER>
        </FILTERS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "checkpoint_filter.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    #test here
    with NekSessionFile(filename) as f:
        checkpoint_filters: list[CheckpointFilter] = f.get_filters("Checkpoint")

    assert len(checkpoint_filters) == 1
    assert checkpoint_filters[0].output_file == "checkpoint_filter.xml"
    assert checkpoint_filters[0].output_frequency == 100
    assert checkpoint_filters[0].output_start_time == 0.0

###############################################################################
# Test get all function files
###############################################################################
@pytest.mark.unit
def test_get_function_files(tmp_path):
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <FUNCTION NAME="InitialConditions">
                <F VAR="u,v,p" FILE="test1.rst" />
            </FUNCTION>
            <FUNCTION NAME="ExactSolution">
                <E VAR="u" VALUE="(1-exp(-0.963740544195769*x)*cos(2*3.141592653589793*y))" />
                <E VAR="v" VALUE="(-0.963740544195769/(2*3.141592653589793))*exp(-0.963740544195769*x)*sin(2*3.141592653589793*y)" />
                <E VAR="p" VALUE="0.5*(1-exp(-2*0.963740544195769*x))" />
            </FUNCTION>
        </CONDITIONS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "function_files.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    with NekSessionFile(filename) as f:
        function_files: list[str] = f.get_all_function_files()

    expected_files: list[str] = ["test1.rst"]

    assert function_files == expected_files

###############################################################################
# Test get expansion functionality
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("test_case",SESSION_EXPANSION_DATASETS,ids=lambda dataset: str(dataset[0]))
def test_expansion_info(test_case: tuple[NekTestDataset,NekTestSessionExpansion]):
    test_dataset,session_expansion_info = test_case

    session: str = test_dataset.session
    geometry: str = test_dataset.geometry
    expected_expansion_data: dict[int,list[ExpansionDefinition]] = session_expansion_info.composite_id_to_expansion_definitions_map

    with NekSessionFile(session) as f:
        expansion_data: dict[int,list[ExpansionDefinition]] = f.get_expansions(geometry)
    
    assert len(expected_expansion_data) == len(expansion_data), f"Expected {len(expected_expansion_data)} expansions, but got {len(expansion_data)}"

    for composite_id in expansion_data.keys():
        assert expansion_data[composite_id] == expansion_data[composite_id], f"Mismatch for composite ID {composite_id}: expected {expected_expansion_data[composite_id]}, but got {expansion_data[composite_id]}"