import h5py
import pytest

from NekUpload.validate.geometry import ValidateGeometry,GeometrySchemaHDF5Validator
from NekUpload.validate.exceptions import HDF5SchemaException,GeometryFileException,HDF5SchemaMissingDatasetException
from NekUpload.validate.exceptions import HDF5SchemaInconsistentException,HDF5SchemaMissingDefinitionException,HDF5SchemaExtraDefinitionException
from NekUpload.testutils import DATASETS,NekTestDataset
###############################################################################
# Tests for GEOMETRY VALIDATOR
#  Test using actual NEKTAR files
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("dataset",DATASETS,ids=str)
def test_check_schema_valid(dataset:NekTestDataset):
    nekg_file: str = dataset.geometry
    checker = ValidateGeometry(nekg_file)
    assert checker.check_schema(),f"{nekg_file} should have passed geometry schema check"

@pytest.mark.unit
@pytest.mark.parametrize("dataset",DATASETS,ids=str)
def test_check_schema_invalid_geometry(dataset:NekTestDataset):
    invalid_nekg_file: str = dataset.output
    checker = ValidateGeometry(invalid_nekg_file)
    with pytest.raises(HDF5SchemaException):
        checker.check_schema()
        pytest.fail(f"Test for {invalid_nekg_file} expected to fail as these are HDF5 output files, not HDF5 geometry files")

@pytest.mark.unit
@pytest.mark.parametrize("dataset",DATASETS,ids=str)
def test_check_schema_invalid_hdf5(dataset:NekTestDataset):
    invalid_nekg_file: str = dataset.session
    checker = ValidateGeometry(invalid_nekg_file)
    with pytest.raises(GeometryFileException):
        checker.check_schema()
        pytest.fail(f"Test for {invalid_nekg_file} expected to fail as these are XML session files, not HDF5 geometry files")

###############################################################################
# Tests for HDF5 GEOMETRY SCHEMA CHECKER
# Test using actual NEKTAR files and use fixtures containing edited versions of
# these files
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("dataset",DATASETS,ids=str)
def test_hdf5_geometry_validator_accept(dataset:NekTestDataset):
    nekg_file: str = dataset.geometry
    with h5py.File(nekg_file) as f:
        validator = GeometrySchemaHDF5Validator(f)
        assert validator.validate(), f"{nekg_file} failed geometry hdf5 validation"

@pytest.mark.unit
@pytest.mark.parametrize("dataset",DATASETS,ids=str)
def test_hdf5_geometry_validator_reject_wrong_output_files(dataset:NekTestDataset):
    geometry_file: str = dataset.output    
    with h5py.File(geometry_file) as f:
        validator = GeometrySchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaException):
            validator.validate()
            pytest.fail(f"{geometry_file} succeeded geometry hdf5 validation. Should fail.")

@pytest.mark.unit
def test_hdf5_geometry_validator_missing_mesh_pair(create_missing_mesh_pair_file):
    with h5py.File(create_missing_mesh_pair_file) as f:
        validator = GeometrySchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaMissingDatasetException):
            validator.validate()
            pytest.fail(f"{create_missing_mesh_pair_file} succeeded geometry hdf5 validation. Should fail as MAPS/HEX exists but MESH/HEX doesn't.")

@pytest.mark.unit
def test_hdf5_geometry_validator_missing_maps_pair(create_missing_maps_pair_file):
    with h5py.File(create_missing_maps_pair_file) as f:
        validator = GeometrySchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaMissingDatasetException):
            validator.validate()
            pytest.fail(f"{create_missing_maps_pair_file} succeeded geometry hdf5 validation. Should fail as MESH/HEX exists but MAPS/HEX doesn't.")

@pytest.mark.unit
def test_hdf5_geometry_validator_inconsistent_definitions(create_missing_inconsistent_pair_file):
    with h5py.File(create_missing_inconsistent_pair_file) as f:
        validator = GeometrySchemaHDF5Validator(f)
        
        with pytest.raises(HDF5SchemaInconsistentException):
            validator.validate()
            pytest.fail(f"{create_missing_inconsistent_pair_file} succeeded geometry hdf5 validation. Should fail as MESH/HEX and MAPS/HEX have inconsistent definitions.")

@pytest.mark.unit
def test_hdf5_geometry_validator_hex_but_no_quad(create_hex_with_missing_2d_file):
    with h5py.File(create_hex_with_missing_2d_file) as f:
        validator = GeometrySchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaMissingDefinitionException):
            validator.validate()
            pytest.fail(f"{create_hex_with_missing_2d_file} succeeded geometry hdf5 validation. Should fail as HEX exist, but no QUADS.")

@pytest.mark.unit
def test_hdf5_geometry_validator_tet_but_no_tri(create_tet_with_missing_2d_file):
    with h5py.File(create_tet_with_missing_2d_file) as f:
        validator = GeometrySchemaHDF5Validator(f)
        
        with pytest.raises(HDF5SchemaMissingDefinitionException):
            validator.validate()  
            pytest.fail(f"{create_tet_with_missing_2d_file} succeeded geometry hdf5 validation. Should fail as TET exist, but no TRIS.")

@pytest.mark.unit
def test_hdf5_geometry_validator_pyr_but_no_quad(create_pyr_with_missing_2d_quad_file):
    with h5py.File(create_pyr_with_missing_2d_quad_file) as f:
        validator = GeometrySchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaMissingDefinitionException):
            validator.validate()  
            pytest.fail(f"{create_pyr_with_missing_2d_quad_file} succeeded geometry hdf5 validation. Should fail as PYR exist, but no QUADS.")

@pytest.mark.unit
def test_hdf5_geometry_validator_pyr_but_no_tris(create_pyr_with_missing_2d_tri_file):
    with h5py.File(create_pyr_with_missing_2d_tri_file) as f:
        validator = GeometrySchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaMissingDefinitionException):
            validator.validate()  
            pytest.fail(f"{create_pyr_with_missing_2d_tri_file} succeeded geometry hdf5 validation. Should fail as PYR exist, but no TRIS.")

@pytest.mark.unit
def test_hdf5_geometry_validator_prism_but_no_quad(create_prism_with_missing_2d_quad_file):
    with h5py.File(create_prism_with_missing_2d_quad_file) as f:
        validator = GeometrySchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaMissingDefinitionException):
            validator.validate()
            pytest.fail(f"{create_prism_with_missing_2d_quad_file} succeeded geometry hdf5 validation. Should fail as PRISM exist, but no QUADS.")

@pytest.mark.unit
def test_hdf5_geometry_validator_prism_but_no_tris(create_prism_with_missing_2d_tri_file):
    with h5py.File(create_prism_with_missing_2d_tri_file) as f:
        validator = GeometrySchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaMissingDefinitionException):
            validator.validate()
            pytest.fail(f"{create_prism_with_missing_2d_tri_file} succeeded geometry hdf5 validation. Should fail as PRISM exist, but no TRIS.")

@pytest.mark.unit
def test_hdf5_geometry_validator_prism_but_no_tris(create_hex_with_insufficient_quads_file):
    with h5py.File(create_hex_with_insufficient_quads_file) as f:
        validator = GeometrySchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaMissingDefinitionException):
            validator.validate()
            pytest.fail(f"{create_hex_with_insufficient_quads_file} succeeded geometry hdf5 validation. Should fail as HEX exist, but insufficeint QUADS.")

@pytest.mark.unit
def test_hdf5_geometry_validator_dangerous_group(create_dangerous_group_file):
    with h5py.File(create_dangerous_group_file) as f:
        validator = GeometrySchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaExtraDefinitionException):
            validator.validate()  
            pytest.fail(f"{create_dangerous_group_file} succeeded geometry hdf5 validation. Should fail as extra groups exist.")

@pytest.mark.unit
def test_hdf5_geometry_validator_multiple_group(create_multiple_dangerous_group_file):
    with h5py.File(create_multiple_dangerous_group_file) as f:
        validator = GeometrySchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaExtraDefinitionException):
            validator.validate()
            pytest.fail(f"{create_multiple_dangerous_group_file} succeeded geometry hdf5 validation. Should fail as extra groups exist.")

@pytest.mark.unit
def test_hdf5_geometry_validator_dangerous_dataset(create_dangerous_dataset_file):
    with h5py.File(create_dangerous_dataset_file) as f:
        validator = GeometrySchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaExtraDefinitionException):
            validator.validate()  
            pytest.fail(f"{create_dangerous_dataset_file} succeeded geometry hdf5 validation. Should fail as extra datasets exist.")

@pytest.mark.unit
def test_hdf5_geometry_validator_multiple_dataset(create_multiple_dangerous_dataset_file):
    with h5py.File(create_multiple_dangerous_dataset_file) as f:
        validator = GeometrySchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaExtraDefinitionException):
            validator.validate()  
            pytest.fail(f"{create_multiple_dangerous_dataset_file} succeeded geometry hdf5 validation. Should fail as extra datasets exist.")