from NekUpload.validate.hdf5_definitions import *
from NekUpload.validate.exceptions import HDF5SchemaException,HDF5SchemaExistenceException
import pytest
import h5py
import numpy as np

###############################################################################
# Set up fixtures here
# Use tmp_path for auto cleanup of files
###############################################################################

@pytest.fixture
def test_file(tmp_path):
    filename = tmp_path / "test.h5"

    with h5py.File(filename,"w") as f:
        nektar = f.create_group("NEKTAR")

        #groups with no attributes
        group1 = nektar.create_group("GROUP1") #i.e. NEKTAR/GROUP1
        group1.create_group("GROUP11")         #NEKTAR/GROUP1/GROUP11
        nektar.create_group("GROUP2")          #NEKTAR/GROUP2

        #add group with attribute
        group3 = nektar.create_group("GROUP3") #NEKTAR/GROUP3
        group3.attrs["ATTRIBUTE"] = "HELLO"

        #add dataset
        data = np.random.rand(10,10)
        nektar.create_dataset("DATASET_1D",data=np.array([10,20,30,40])) # NEKTAR/DATASET_1D
        nektar.create_dataset("DATASET_2D",data=data) # NEKTAR/DATASET_2D

    return filename

###############################################################################
# TEST GROUP VALIDATOR
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("hdf5_group,attributes_list",[
    ("NEKTAR",[]),
    ("NEKTAR/GROUP1",[]),
    ("NEKTAR/GROUP1/GROUP11",[]),
    ("NEKTAR/GROUP2",[]),
    ("NEKTAR/GROUP3",["ATTRIBUTE"])
])
def test_valid_hdf5_group(test_file,hdf5_group:str,attributes_list:list[str]):
    group_def = HDF5GroupDefinition(hdf5_group,attributes_list)
        
    with h5py.File(test_file,"r") as f:
        assert group_def.validate(f), f"Validation failed for group '{hdf5_group}' with attributes {attributes_list}"

@pytest.mark.unit
@pytest.mark.parametrize("description,hdf5_group,attributes_list,exception",[
    ("non-existent-group","NEKTAR/NO/EXISTO",[],HDF5SchemaExistenceException),
    ("non-existent-attr","NEKTAR/GROUP1",["ATTRSNOEXIST"],HDF5SchemaInconsistentException),
    ("non-existent-attr2","NEKTAR/GROUP3",["ATTRIBUTE","ATTRSNOEXIST"],HDF5SchemaInconsistentException),
    ("is-dataset","NEKTAR/DATASET_1D",[],HDF5SchemaInconsistentException),
    ("is-dataset2","NEKTAR/DATASET_2D",["ATTRIBUTE"],HDF5SchemaInconsistentException),
])
def test_hdf5_group_wrong(test_file,description:str,hdf5_group:str,attributes_list:list[str],exception:Exception):
    group_def = HDF5GroupDefinition(hdf5_group,attributes_list)

    with h5py.File(test_file, "r") as f:
        with pytest.raises(exception):
            group_def.validate(f)
            pytest.fail(f"Test failed: {description} - Expected exception for group '{hdf5_group}', but none was raised.")

###############################################################################
# TEST DATASET VALIDATOR
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("description,hdf5_dataset,dataset_shape",[
    ("no-constraints","NEKTAR/DATASET_1D",()),
    ("no-constraints","NEKTAR/DATASET_2D",()),
    ("1d-dataset-constraints","NEKTAR/DATASET_1D",(-1,)), #-1 is wildcard, i.e. no real constraint, just that it is 1d
    ("2d-dataset-constraints","NEKTAR/DATASET_2D",(-1,10)), #as defined in fixture
])
def test_valid_hdf5_dataset(test_file,description:str,hdf5_dataset:str,dataset_shape:tuple[int,...]):
    dataset_def = HDF5DatasetDefinition(hdf5_dataset,dataset_shape)

    with h5py.File(test_file,"r") as f:
        assert dataset_def.validate(f), f"Test {description}: Validation failed for dataset '{hdf5_dataset}' with shape constraint {dataset_shape}"

@pytest.mark.unit
@pytest.mark.parametrize("description,hdf5_dataset,dataset_shape,exception",[
    ("nonexistent-dataset","NEKTAR/DATASET/NON/EXISTENT",(),HDF5SchemaExistenceException),
    ("invalid-dimension-mismatched-constraints","NEKTAR/DATASET_1D",(-1,5),HDF5SchemaInconsistentException),
    ("invalid-second-dimension-constraints","NEKTAR/DATASET_2D",(-1,2),HDF5SchemaInconsistentException),
    ("group-not-dataset","NEKTAR/GROUP1",(),HDF5SchemaInconsistentException)
])
def test_invalid_hdf5_dataset(test_file,description:str,hdf5_dataset:str,dataset_shape:tuple[int,...],exception:Exception):
    dataset_def = HDF5DatasetDefinition(hdf5_dataset,dataset_shape)

    with h5py.File(test_file,"r") as f:
        with pytest.raises(exception):
            dataset_def.validate(f)
            pytest.fail(f"Test {description}: Validation did not fail for dataset '{hdf5_dataset}' with shape constraint {dataset_shape}")