import h5py
import pytest

from NekUpload.validate.output import ValidateOutput,OutputSchemaHDF5Validator,HDF5SchemaInconsistentException
from NekUpload.validate.exceptions import HDF5SchemaException,OutputFileException,HDF5SchemaExtraDefinitionException

from NekUpload.testutils import DATASETS,NekTestDataset
###############################################################################
# Test Schema Validation with real nektar files
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("dataset",DATASETS,ids=str)
def test_check_output_schema_valid(dataset:NekTestDataset):
    fld_files:str = dataset.output
    checker = ValidateOutput(fld_files)
    assert checker.check_schema(), f"Schema check failed for {fld_files} when it should have succeeded"

@pytest.mark.unit
@pytest.mark.parametrize("dataset",DATASETS,ids=str)
def test_check_schema_invalid_output(dataset:NekTestDataset):
    invalid_fld_file: str = dataset.geometry #still hdf5 files, but not an output file
    
    checker = ValidateOutput(invalid_fld_file)
    with pytest.raises(HDF5SchemaException):
        checker.check_schema()
        pytest.fail(f"Test for {invalid_fld_file} should have failed, these are HDF5 geometry files, not HDF5 Output files")

@pytest.mark.unit
@pytest.mark.parametrize("dataset",DATASETS,ids=str)
def test_check_schema_invalid_hdf5(dataset:NekTestDataset):
    invalid_fld_file: str = dataset.session #these are no longer hdf5 files
    checker = ValidateOutput(invalid_fld_file)
    with pytest.raises(OutputFileException):
        checker.check_schema()
        pytest.fail(f"Test {invalid_fld_file} failed to raise error, these are XML session files, not HDF5 Output files")

###############################################################################
# Test Output HDF5 Schema Checker
# Use real nektar files adn set of fixtures containing various invalid 
# output files in various ways
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("dataset",DATASETS,ids=str)
def test_hdf5_output_validator_accept(dataset:NekTestDataset):
    fld_file: str = dataset.output
    with h5py.File(fld_file) as f:
        validator = OutputSchemaHDF5Validator(f)
        assert validator.validate(),f"Test {fld_file} is valid output file but failed schema validation"

@pytest.mark.unit
@pytest.mark.parametrize("dataset",DATASETS,ids=str)
def test_hdf5_output_validator_reject_wrong_files(dataset:NekTestDataset):
    nekg_files: str = dataset.geometry
    with h5py.File(nekg_files) as f:
        validator = OutputSchemaHDF5Validator(f)
        
        with pytest.raises(HDF5SchemaException):
            validator.validate()
            pytest.fail(f"Test {nekg_files} is a geometry  file so should fail schema validation")

@pytest.mark.unit
def test_hdf5_output_validator_reject_extra_datasets(create_output_dangerous_datasets_file):
    with h5py.File(create_output_dangerous_datasets_file) as f:
        validator = OutputSchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaExtraDefinitionException):
            validator.validate()        
            pytest.fail(f"{create_output_dangerous_datasets_file} succeeded output hdf5 validation. Should fail as extra datasets present.")

@pytest.mark.unit
def test_hdf5_output_validator_reject_one_extra_dataset(create_output_one_dangerous_datasets_file):
    with h5py.File(create_output_one_dangerous_datasets_file) as f:
        validator = OutputSchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaExtraDefinitionException):
            validator.validate()        
            pytest.fail(f"{create_output_one_dangerous_datasets_file} succeeded output hdf5 validation. Should fail as extra datasets present.")

@pytest.mark.unit
def test_hdf5_output_validator_reject_extra_groups(create_output_dangerous_groups_file):
    with h5py.File(create_output_dangerous_groups_file) as f:
        validator = OutputSchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaExtraDefinitionException):
            validator.validate()        
            pytest.fail(f"{create_output_dangerous_groups_file} succeeded output hdf5 validation. Should fail as extra groups present.")

@pytest.mark.unit
def test_hdf5_output_validator_reject_one_extra_groups(create_output_one_dangerous_groups_file):
    with h5py.File(create_output_one_dangerous_groups_file) as f:
        validator = OutputSchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaExtraDefinitionException):
            validator.validate()        
            pytest.fail(f"{create_output_one_dangerous_groups_file} succeeded output hdf5 validation. Should fail as extra groups present.")

@pytest.mark.unit
def test_hdf5_output_validator_decomposition_1d_shape_incorrect(create_incorrect_decomposition_1d_shape_file):
    with h5py.File(create_incorrect_decomposition_1d_shape_file) as f:
        validator = OutputSchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaInconsistentException):
            validator.validate()        
            pytest.fail(f"{create_incorrect_decomposition_1d_shape_file} succeeded output hdf5 validation. Should fail as decomposition has 13 elements, not multiple of 7.")

@pytest.mark.unit
def test_hdf5_output_validator_decomposition_2d_shape_incorrect(create_incorrect_decomposition_2d_shape_file):
    with h5py.File(create_incorrect_decomposition_2d_shape_file) as f:
        validator = OutputSchemaHDF5Validator(f)

        with pytest.raises(HDF5SchemaInconsistentException):
            validator.validate()        
            pytest.fail(f"{create_incorrect_decomposition_2d_shape_file} succeeded output hdf5 validation. Should fail as decomposition has 14x2 elements, not 14x1.")