import os
import pytest
import h5py
import numpy as np

from NekUpload.validate.session import ValidateSession
from NekUpload.validate.exceptions import MissingInputFileException,MissingOutputFileException,SessionFileException
from NekUpload.testutils import BOUNDARY_CONDITION_DATASETS,DATASETS,FUNCTIONS_DATASETS,NekTestDataset

@pytest.mark.unit
@pytest.mark.parametrize("dataset",DATASETS,ids=str)
def test_nektar_schema(dataset:NekTestDataset,nektar_session_schema):
    session: str = dataset.session    
    nektar_schema: str = nektar_session_schema

    checker = ValidateSession(session)
    assert checker.is_valid_xml(session,nektar_schema), f"File {session} failed schema {nektar_schema} checks"

@pytest.mark.unit
@pytest.mark.parametrize("bc_dataset",BOUNDARY_CONDITION_DATASETS,ids=str)
def test_nektar_bc_schema(bc_dataset:NekTestDataset,nektar_bc_schema):
    session: str = bc_dataset.session
    boundary_condition: list[str] = bc_dataset.boundary_conditions
    nektar_bc_schema: str = nektar_bc_schema

    checker = ValidateSession(session)

    for bc_file in boundary_condition:
        assert checker.is_valid_xml(bc_file, nektar_bc_schema), f"Boundary condition file {bc_file} failed schema {nektar_bc_schema} checks"

@pytest.mark.unit
@pytest.mark.parametrize("dataset",DATASETS,ids=str)
def test_check_schema(dataset:NekTestDataset):
    session: str = dataset.session
    
    checker = ValidateSession(session)
    assert checker.check_schema(),f"Validation failed for XML file: {session}"

@pytest.mark.unit
@pytest.mark.parametrize("dataset",FUNCTIONS_DATASETS,ids=str)
def test_check_function_files(dataset:NekTestDataset):
    session: str = dataset.session
    function_files: list[str] = dataset.functions

    checker = ValidateSession(session)
    assert checker.check_function_reference(list(function_files))

@pytest.mark.unit
@pytest.mark.parametrize("dataset",FUNCTIONS_DATASETS,ids=str)
def test_check_filter_output_complete(dataset:NekTestDataset):
    session: str = dataset.session
    files: list[str] = dataset.checkpoints + dataset.filters

    checker = ValidateSession(session)
    assert checker.check_filter_files_reference(list(files))
    
###############################################################################
# Test Boundary Condition
###############################################################################

@pytest.fixture
def write_valid_boundary_conditions_files(tmp_path):
    """Only need bare minimum for this
    Geometry -> Needs NEKTAR/GEOMETRY/MAPS/COMPOSITE and MESH/COMPOSITE
    Session -> NEKTAR/CONDITIONS/BOUNDARYREGIONS and BOUNDARYCONDITIONS and NEKTAR/GEOMETRY
    """
    geometry_file = str(tmp_path / "geometryBC.nekg")
    session_file = str(tmp_path / "sessionBC.xml")

    with h5py.File(geometry_file,"w") as f:
        root = f.create_group("NEKTAR")
        geometry = root.create_group("GEOMETRY")
        maps_group = geometry.create_group("MAPS")
        mesh_group = geometry.create_group("MESH")

        mesh_composite_data = np.array([
            b' Q[22-47] ', b' T[0-21] ', b' E[0-1] ', b' E[2-5] '
        ], dtype="S")  # "S" ensures it's stored as a byte string

        maps_composite_data = np.arange(4, dtype=np.int32)  # Integer array [0, 1, ..., 10]
        mesh_group.create_dataset("COMPOSITE", data=mesh_composite_data)
        maps_group.create_dataset("COMPOSITE", data=maps_composite_data)

    xml_content = '''<NEKTAR>
        <GEOMETRY DIM="2" SPACE="2" HDF5FILE="TEST.nekg"/>
        <CONDITIONS>
            <BOUNDARYREGIONS>
                <B ID="0">C[2]</B>
                <B ID="1">C[3]</B>
            </BOUNDARYREGIONS>
            
            <BOUNDARYCONDITIONS>
                <REGION REF="0">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="1">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
            </BOUNDARYCONDITIONS>
        </CONDITIONS>
    </NEKTAR>
    '''    
    
    # Write to a file
    with open(session_file, 'w') as file:
        file.write(xml_content)

    return geometry_file,session_file

@pytest.fixture
def write_boundary_condition_wrong_dim_files(tmp_path):
    session_file = str(tmp_path / "sessionBC.xml")
    xml_content = '''<NEKTAR>
        <GEOMETRY DIM="3" SPACE="3" HDF5FILE="TEST.nekg"/>
        <CONDITIONS>
            <BOUNDARYREGIONS>
                <B ID="0">C[2]</B>
                <B ID="1">C[3]</B>
            </BOUNDARYREGIONS>
            
            <BOUNDARYCONDITIONS>
                <REGION REF="0">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="1">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
            </BOUNDARYCONDITIONS>
        </CONDITIONS>
    </NEKTAR>
    '''    
    
    # Write to a file
    with open(session_file, 'w') as file:
        file.write(xml_content)

    return session_file

@pytest.fixture
def write_boundary_condition_wrong_composites_defined_files(tmp_path):
    session_file = str(tmp_path / "sessionBC.xml")
    xml_content = '''<NEKTAR>
        <GEOMETRY DIM="2" SPACE="2" HDF5FILE="TEST.nekg"/>
        <CONDITIONS>
            <BOUNDARYREGIONS>
                <B ID="0">C[2]</B>
                <B ID="1">C[4]</B>
            </BOUNDARYREGIONS>
            
            <BOUNDARYCONDITIONS>
                <REGION REF="0">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
                <REGION REF="1">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                </REGION>
            </BOUNDARYCONDITIONS>
        </CONDITIONS>
    </NEKTAR>
    '''    
    
    # Write to a file
    with open(session_file, 'w') as file:
        file.write(xml_content)

    return session_file

@pytest.mark.unit
def test_check_valid_full_boundary_condition_definition(write_valid_boundary_conditions_files):
    geometry_file,session_file = write_valid_boundary_conditions_files

    validator = ValidateSession(session_file)
    assert validator.check_boundary_definition(geometry_file), f"Boundary condition definition check failed for geometry file: {geometry_file} in session: {session_file}"

@pytest.mark.unit
def test_check_invalid_full_boundary_condition_definition(write_valid_boundary_conditions_files,write_boundary_condition_wrong_dim_files):
    geometry_file,_ = write_valid_boundary_conditions_files
    session_file = write_boundary_condition_wrong_dim_files

    validator = ValidateSession(session_file)
    with pytest.raises(SessionFileException):
        validator.check_boundary_definition(geometry_file)

@pytest.mark.unit
def test_check_wrong_composites_boundary_condition_definition(write_valid_boundary_conditions_files,write_boundary_condition_wrong_composites_defined_files):
    geometry_file,_ = write_valid_boundary_conditions_files
    session_file = write_boundary_condition_wrong_composites_defined_files

    validator = ValidateSession(session_file)
    with pytest.raises(SessionFileException):
        validator.check_boundary_definition(geometry_file)

@pytest.mark.integration
def test_check_full_boundary_condition_definition(ADR_dataset_abs_paths):
    datasets: list[dict[str,str | list[str]]] = ADR_dataset_abs_paths

    for dataset in datasets:
        session: str = dataset["SESSION"]
        geometry: str = dataset["GEOMETRY"]

        validator = ValidateSession(session)
        assert validator.check_boundary_definition(geometry), f"Boundary condition definition check failed for geometry file: {geometry} in session: {session}"

###############################################################################
# Test correct geometry file references
###############################################################################

@pytest.mark.unit
def test_valid_geometry_file_reference(tmp_path):
    #test setup 
    session_file = str(tmp_path / "sessionGeometry.xml")
    xml_content = '''<NEKTAR>
        <GEOMETRY DIM="2" SPACE="2" HDF5FILE="TEST.nekg"/>
    </NEKTAR>
    '''    
    
    # Write to a file
    with open(session_file, 'w') as file:
        file.write(xml_content)

    #Actual test
    validator = ValidateSession(session_file)
    assert validator.check_geometry_file_reference("TEST.nekg"), f"Geometry file reference check failed for session file: {session_file}"

@pytest.mark.unit
def test_invalid_geometry_file_reference(tmp_path):
    #test setup 
    session_file = str(tmp_path / "sessionGeometry.xml")
    xml_content = '''<NEKTAR>
        <GEOMETRY DIM="2" SPACE="2" HDF5FILE="INVALID.nekg"/>
    </NEKTAR>
    '''    
    
    # Write to a file
    with open(session_file, 'w') as file:
        file.write(xml_content)

    #Actual test
    validator = ValidateSession(session_file)
    with pytest.raises(SessionFileException):
        validator.check_geometry_file_reference("SOMETHINGELSE.nekg")

###############################################################################
# Test whether boundary conditions paths are correctly detected
###############################################################################

@pytest.fixture
def write_bc_reference_file(tmp_path):
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <BOUNDARYCONDITIONS>
                <REGION REF="0">
                    <D VAR="u" FILE="test1.bc" />
                </REGION>
                <REGION REF="1">
                    <D VAR="u" USERDEFINEDTYPE="TimeDependent" VALUE="sin(PI*(x-advx*t))*sin(PI*(y-advy*t))" />
                    <D VAR="v" FILE="test2.bc" />
                </REGION>
            </BOUNDARYCONDITIONS>
        </CONDITIONS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "geometry_info_only.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    return filename

@pytest.mark.unit
def test_valid_bc_files(write_bc_reference_file):
    file_list = ["test1.bc","file.jpg","test2.bc"]

    validator = ValidateSession(write_bc_reference_file)
    assert validator.check_boundary_conditions_reference(file_list), f"Boundary conditions reference check failed for file list: {file_list}"

@pytest.mark.unit
def test_invalid_bc_files(write_bc_reference_file):
    file_list = ["test1.bc","file.jpg","test2.jpg"]

    validator = ValidateSession(write_bc_reference_file)
    with pytest.raises(SessionFileException):
        validator.check_boundary_conditions_reference(file_list)

###############################################################################
# Test whether boundary condition files are identified and verified
###############################################################################

@pytest.mark.integration
def test_valid_bc_schema_check(dataset_with_boundary_conditions_file):
    session_path,_,boundary_condition_path,_ = dataset_with_boundary_conditions_file

    validator = ValidateSession(session_path)
    validator.check_boundary_condition_schema([boundary_condition_path])

###############################################################################
# Test the checkpoint file identifiers and associated helper functions
###############################################################################

@pytest.mark.unit
@pytest.mark.parametrize("num_step,time_step,final_time,expected_num_step,expected_time_step,expected_final_time",[
    (100,0.01,1,100,0.01,1),
    (100,0.01,None,100,0.01,1),
    (100,None,1,100,0.01,1),
    (None,0.01,1,100,0.01,1),
])
def test_valid_resolution_of_timestep_params(tmp_path,num_step:int,time_step:float,
                                            final_time:float,expected_num_step:int,
                                            expected_time_step:float,expected_final_time:float):
    filename = str(tmp_path / "dummy_file.xml")
    with open(filename, 'w') as file:
        file.write("<NEKTAR>PLACEHOLDER TEST FILE</NEKTAR>")

    validator = ValidateSession(filename)
    num_step,time_step,final_time = validator._resolve_timestep_parameters(num_step,time_step,final_time)

    assert num_step == pytest.approx(expected_num_step,1e-6), f"Expected num_step to be {expected_num_step}, but got {num_step}"
    assert time_step == pytest.approx(expected_time_step,1e-6), f"Expected time_step to be {expected_time_step}, but got {time_step}"
    assert final_time == pytest.approx(expected_final_time,1e-6), f"Expected final_time to be {expected_final_time}, but got {final_time}"

@pytest.mark.unit
@pytest.mark.parametrize("num_step,time_step,final_time",[
    (100,None,None),
    (None,None,1),
    (None,0.01,None),
    (100,0.01,10)#this one all defined, but mismatched
])
def test_invalid_resolution_of_timestep_params(tmp_path,num_step:int,time_step:float,final_time:float):
    filename = str(tmp_path / "dummy_file.xml")
    with open(filename, 'w') as file:
        file.write("<NEKTAR>PLACEHOLDER TEST FILE</NEKTAR>")

    validator = ValidateSession(filename)
    with pytest.raises(SessionFileException):
        validator._resolve_timestep_parameters(num_step,time_step,final_time)

@pytest.fixture
def write_session_generates_only_one_initial_chk_file(tmp_path):
    """This should only give a .chk0"""
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <PARAMETERS>
                <P>FinTime = 0.01</P>
                <P>TimeStep = 0.001</P>
                <P>NumSteps = FinTime/TimeStep</P>
                <P>IO_CheckSteps = 999</P>
                <P>IO_InfoSteps = 10</P>
            </PARAMETERS>
        </CONDITIONS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "chk_info.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    return filename

@pytest.fixture
def write_session_generates_no_chk_file(tmp_path):
    """This should give no .chk0"""
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <PARAMETERS>
                <P>Lambda = 1</P>
                <P>IterativeSolverTolerance = 1e-8</P>
            </PARAMETERS>
        </CONDITIONS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "chk_info.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    return filename

@pytest.fixture
def write_session_generates_no_chk_file2(tmp_path):
    """This should give no .chk0"""
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <PARAMETERS>
                <P>TimeStep = 0.0</P>
                <P>NumSteps = 100</P>
                <P>IO_InfoSteps = 100</P>
                <P>IO_CheckSteps = 0</P>
            </PARAMETERS>
        </CONDITIONS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "chk_info.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    return filename

@pytest.fixture
def write_session_generates_two_chk_files(tmp_path):
    """This should give a .chk0 and .chk1"""
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <PARAMETERS>
                <P>NumSteps = 100</P>
                <P>TimeStep = 0.001</P>
                <P>IO_CheckSteps = 100</P>
                <P>IO_InfoSteps = 50</P>
            </PARAMETERS>
        </CONDITIONS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "chk_info.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    return filename

@pytest.mark.unit
def test_compute_correct_chk_info_from_parameters_initial_file_only(write_session_generates_only_one_initial_chk_file):
    session_file = write_session_generates_only_one_initial_chk_file

    expected_chk_file_num = 1
    expected_chk_file_times = (0,)

    validator = ValidateSession(session_file)
    chk_file_num,file_times = validator._compute_chk_file_requirements_from_parameters()
    
    assert chk_file_num == expected_chk_file_num, f"Expected checkpoint file number to be {expected_chk_file_num}, but got {chk_file_num}"
    assert file_times == expected_chk_file_times, f"Expected checkpoint file times to be {expected_chk_file_times}, but got {file_times}"

@pytest.mark.unit
def test_compute_correct_chk_info_from_parameters_no_chk_files(write_session_generates_no_chk_file):
    session_file = write_session_generates_no_chk_file

    expected_chk_file_num = 0
    expected_chk_file_times = ()

    validator = ValidateSession(session_file)
    chk_file_num,file_times = validator._compute_chk_file_requirements_from_parameters()
    
    assert chk_file_num == expected_chk_file_num, f"Expected checkpoint file number to be {expected_chk_file_num}, but got {chk_file_num}"
    assert file_times == expected_chk_file_times, f"Expected checkpoint file times to be {expected_chk_file_times}, but got {file_times}"

@pytest.mark.unit
def test_compute_correct_chk_info_from_parameters_no_chk_files_2(write_session_generates_no_chk_file2):
    session_file = write_session_generates_no_chk_file2

    expected_chk_file_num = 0
    expected_chk_file_times = ()

    validator = ValidateSession(session_file)
    chk_file_num,file_times = validator._compute_chk_file_requirements_from_parameters()
    
    assert chk_file_num == expected_chk_file_num, f"Expected checkpoint file number to be {expected_chk_file_num}, but got {chk_file_num}"
    assert file_times == expected_chk_file_times, f"Expected checkpoint file times to be {expected_chk_file_times}, but got {file_times}"

@pytest.mark.unit
def test_compute_correct_chk_info_from_parameters_no_chk_files_2(write_session_generates_two_chk_files):
    session_file = write_session_generates_two_chk_files

    expected_chk_file_num = 2
    expected_chk_file_times = (0,0.1)

    validator = ValidateSession(session_file)
    chk_file_num,file_times = validator._compute_chk_file_requirements_from_parameters()
    
    assert chk_file_num == expected_chk_file_num, f"Expected checkpoint file number to be {expected_chk_file_num}, but got {chk_file_num}"
    assert file_times == expected_chk_file_times, f"Expected checkpoint file times to be {expected_chk_file_times}, but got {file_times}"

@pytest.fixture
def write_session_generates_only_one_initial_chk_filter_file(tmp_path):
    """This should only give a .chk0"""
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <PARAMETERS>
                <P>FinTime = 0.01</P>
                <P>TimeStep = 0.001</P>
                <P>NumSteps = FinTime/TimeStep</P>
            </PARAMETERS>
        </CONDITIONS>
        <FILTERS>
            <FILTER TYPE="Checkpoint"> 
                <PARAM NAME="OutputFrequency">999</PARAM> 
            </FILTER>
        </FILTERS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "chk_info.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    return filename

@pytest.fixture
def write_session_generates_only_two_initial_chk_filter_file(tmp_path):
    """This should only give a .chk0 and .chk1"""
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <PARAMETERS>
                <P>NumSteps = 100</P>
                <P>TimeStep = 0.001</P>
                <P>IO_InfoSteps = 50</P>
            </PARAMETERS>
        </CONDITIONS>
        <FILTERS>
            <FILTER TYPE="Checkpoint"> 
                <PARAM NAME="OutputFrequency">100</PARAM> 
            </FILTER>
        </FILTERS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "chk_info.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    return filename

@pytest.fixture
def write_session_generates_only_two_initial_chk_filter_file_diff_start(tmp_path):
    """This should only give a .chk0 and .chk1"""
    xml_content = '''<NEKTAR>
        <CONDITIONS>
            <PARAMETERS>
                <P>NumSteps = 100</P>
                <P>TimeStep = 0.001</P>
                <P>IO_InfoSteps = 50</P>
            </PARAMETERS>
        </CONDITIONS>
        <FILTERS>
            <FILTER TYPE="Checkpoint"> 
                <PARAM NAME="OutputFrequency">50</PARAM> 
                <PARAM NAME="OutputStartTime">0.05</PARAM> 
            </FILTER>
        </FILTERS>
    </NEKTAR>
    '''
    # Write to a file
    filename = str(tmp_path / "chk_info.xml")
    with open(filename, 'w') as file:
        file.write(xml_content)

    return filename

@pytest.mark.unit
def test_compute_correct_chk_info_from_filters(write_session_generates_only_one_initial_chk_filter_file):
    session_file = write_session_generates_only_one_initial_chk_filter_file

    expected_chk_file_num = 1
    expected_chk_file_times = (0,)

    validator = ValidateSession(session_file)
    chk_file_num,file_times = validator._compute_chk_file_requirements_from_filter()
    
    assert chk_file_num == expected_chk_file_num, f"Expected checkpoint file number to be {expected_chk_file_num}, but got {chk_file_num}"
    assert file_times == expected_chk_file_times, f"Expected checkpoint file times to be {expected_chk_file_times}, but got {file_times}"

@pytest.mark.unit
def test_compute_correct_chk_info_from_filters_2(write_session_generates_only_two_initial_chk_filter_file):
    session_file = write_session_generates_only_two_initial_chk_filter_file

    expected_chk_file_num = 2
    expected_chk_file_times = (0,0.1)

    validator = ValidateSession(session_file)
    chk_file_num,file_times = validator._compute_chk_file_requirements_from_filter()
    
    assert chk_file_num == expected_chk_file_num, f"Expected checkpoint file number to be {expected_chk_file_num}, but got {chk_file_num}"
    assert file_times == expected_chk_file_times, f"Expected checkpoint file times to be {expected_chk_file_times}, but got {file_times}"

@pytest.mark.unit
def test_compute_correct_chk_info_from_filters_3(write_session_generates_only_two_initial_chk_filter_file_diff_start):
    session_file = write_session_generates_only_two_initial_chk_filter_file_diff_start

    expected_chk_file_num = 2
    expected_chk_file_times = (0.05,0.1)

    validator = ValidateSession(session_file)
    chk_file_num,file_times = validator._compute_chk_file_requirements_from_filter()
    
    assert chk_file_num == expected_chk_file_num, f"Expected checkpoint file number to be {expected_chk_file_num}, but got {chk_file_num}"
    assert file_times == expected_chk_file_times, f"Expected checkpoint file times to be {expected_chk_file_times}, but got {file_times}"

@pytest.mark.integration
def test_compute_chk_file(ADR_dataset_abs_paths):
    datasets = ADR_dataset_abs_paths

    for dataset in datasets:
        session:str = dataset["SESSION"]
        checkpoint: list[str] = dataset["CHECKPOINT"]

        validator = ValidateSession(session)
        assert validator.check_checkpoint_files(checkpoint)

@pytest.mark.integration
def test_compute_chk_file_invalid(ADR_dataset_abs_paths):
    datasets = ADR_dataset_abs_paths

    for dataset in datasets:
        session:str = dataset["SESSION"]
        checkpoint: list[str] = dataset["CHECKPOINT"]
        checkpoint.append("invalid.chk") #add an extra file

        validator = ValidateSession(session)
        with pytest.raises(SessionFileException):
            validator.check_checkpoint_files(checkpoint)